# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""query manager helper functions for use with IPython/Juptyer queries."""
import sys
from functools import partial
import re

from . query_schema import DataSchema
from . query_builtin_queries import query_definitions
from . query_defns import KqlQuery, QueryParamProvider, DataFamily, DataEnvironment
from . utility import export
from .. _version import VERSION

__version__ = VERSION
__author__ = 'Ian Hellen'

# module constants
_DATA_FAMILY_NAME = 'data_family'
_DATA_ENVIRONMENT_NAME = 'data_environment'


# utility functions
@export
def print_kql(query_string: str):
    """Print kql query stripped of comments and newline characters."""
    clean_qry = re.sub(r'(//[^\"\'\n]+)', ' ', query_string, re.MULTILINE).strip()
    for line in clean_qry.split('\n'):
        print(line.strip())


@export
def clean_kql_query(query_string: str) -> str:
    """Return kql query stripped of comments and newline characters."""
    remove_comments = re.sub(r'(//[^\"\'\n]+)', ' ', query_string, re.MULTILINE).strip()
    # get rid of newlines and returns
    return re.sub(r'(\s*\n\s*)', ' ', remove_comments)


@export
def query_help(queryname: str):
    """Print query usage."""
    if queryname not in query_definitions:
        print('Unknown query: ', queryname)
        return

    kql_query = query_definitions[queryname]
    print('Query: ', queryname)
    print(query_definitions[queryname].description)
    print('Designed to be executed with data_source: ',
          kql_query.data_source)
    print('Supported data families: ',
          ', '.join([str(fam) for fam in kql_query.data_families]))
    print('Supported data environments: ',
          ', '.join([str(env) for env in kql_query.data_environments]))

    req_params = required_params(kql_query.query)
    req_params.remove('table')
    req_params.remove('query_project')

    print('Query parameters:')
    print(req_params)
    if kql_query.optional_params:
        print('Optional parameters:')
        print(', '.join([str(param) for param in kql_query.optional_params]))
    print('Query:')
    print_kql(kql_query.query)


@export
def add_query(kql_query: KqlQuery = None, **kwargs):
    """
    Add a query to the current set.

        :param kql_query:KqlQuery:
    """
    if kql_query is None:
        def_data_families = [DataEnvironment.LogAnalytics]
        def_data_environments = [
            DataFamily.WindowsSecurity, DataFamily.LinuxSecurity]
        if 'name' not in kwargs or 'query' not in kwargs or 'data_source' not in kwargs:
            raise ValueError(
                'If kql_query is not supplied the kwargs must include name, query and data_source.')
        kql_query = KqlQuery(name=kwargs['name'],
                             query=kwargs['query'],
                             description=kwargs.get('description', None),
                             data_source=kwargs['data_source'],
                             data_families=kwargs.get(
                                 'data_families', def_data_families),
                             data_environments=kwargs.get('data_environments',
                                                          def_data_environments))
    query_definitions[kql_query.name] = kql_query
    _add_queries_to_module(__name__)

    kql_modules = [m for m in sys.modules if m.endswith('msticpy.nbtools.kql')]
    if len(kql_modules) == 1:
        _add_queries_to_module(kql_modules[0])


@export
def list_queries():
    """Return list of currently defined queries."""
    return list(query_definitions.keys())


@export
def replace_query_params(query_name: str, *args, **kwargs) -> str:
    """
    Return the parameterized query for query_name.

    Arguments:
        query_name {string}: The query to use
        args {QueryParamProvider}: objects that
                implement QueryParamProvider (from which query
                parameters can be extracted).
        kwargs:
            {string:[QueryParamProvider]} - for the key 'provs'
                this should be a collection of objects that
                implement QueryParamProvider (from which query
                parameters can be extracted).
                OR
            {string:value pairs} -- custom parameter list
                (override default values and values extracted
                from QueryParamProviders).
    Raises:
        LookupError -- query_name cannot be found

    Returns:
        string -- substituted query

    """
    return replace_prov_query_params(query_name=query_name,
                                     provs=args,
                                     **kwargs)


@export
def replace_prov_query_params(query_name: str, **kwargs) -> str:
    """
    Return the parameterized query for query_name.

    Arguments:
        query_name {string} -- The query to use
        kwargs
            {string:[QueryParamProvider]} - for the key 'provs'
                this should be a collection of objects that
                implement QueryParamProvider (from which query
                parameters can be extracted).
                OR
            {string:value pairs} -- custom parameter list
                (override default values and values extracted
                from QueryParamProviders).
    Raises:
        LookupError -- query_name cannot be found.
        ValueError -- query parameter value could not be found.

    Returns:
        string -- substituted query

    """
    if query_name not in query_definitions:
        raise LookupError(f'Unknown query "{query_name}"')

    kql_query = query_definitions[query_name]
    if 'provs' in kwargs:
        p_args = kwargs.pop('provs')
        query_params = _get_query_params(kql_query, *p_args, **kwargs)
    else:
        query_params = _get_query_params(kql_query, **kwargs)
    return kql_query.query.format(**query_params)


def _get_query_params(kql_query, *args, **kwargs):
    """
    Get the parameters needed for the query.

    Arguments:
        kql_query {KqlQuery} -- query object
        args {list} -- set of source objects to extract parameter
            values from
        kwargs {string:value pairs} -- custom parameter list
            (overrides auto-extracted values)

    Raises:
        LookupError -- Could not find valid data_family or
            data_environment
        ValueError -- Values for one or more required parameters
            could not be found

    Returns:
        dict -- Dictionary of parameter names and values to be used
            in the query

    """
    # get the required parameters for this query and build a dictionary
    req_param_names = required_params(kql_query.query)
    req_params = {param: None for param in req_param_names}

    # Iterate through required parameters. If any are set in the supplied
    # provider objects, assign them to our output dictionary
    query_providers = [prov for prov in args if isinstance(prov, QueryParamProvider)]
    for provider in query_providers:
        for param in req_param_names:
            if param in provider.query_params:
                req_params[param] = provider.query_params[param]

    # If any custom parameters have been supplied add these
    # overriding any parameters from the QueryParamProviders
    if kwargs:
        req_params.update(kwargs)

    data_family, data_environment = _get_data_family_and_env(kql_query, query_providers, kwargs)

    if not data_family:
        supp_families = ', '.join(DataSchema.get_data_families())
        raise LookupError('Could not find a valid data_family value. '
                          f'Valid families are: {supp_families}')
    if not data_environment:
        supp_envs = ', '.join(DataSchema.get_data_environments())
        raise LookupError('Could not find a valid data_environment value. '
                          f'Valid environments are: {supp_envs}')

    # Create the data schema and get any unset parameters from
    # the data schema
    data_schema = DataSchema(environment=data_environment,
                             data_family=data_family,
                             data_source=kql_query.data_source)

    for param, value in req_params.items():
        if not value and param in data_schema:
            req_params[param] = data_schema[param]

    # If we have missing parameters try to retrieve them
    # as attributes of the object
    missing_params = [p_name for p_name, p_value in req_params.items() if not p_value]
    if missing_params:
        for other_object in [obj for obj in args if not isinstance(obj, QueryParamProvider)]:
            for m_param in missing_params:
                if m_param in other_object:
                    req_params[m_param] = getattr(other_object, m_param)
            missing_params = [p_name for p_name, p_value in req_params.items() if not p_value]

        if missing_params:
            # check for and remove optional parameters from the missing params list
            for m_param in missing_params:
                if m_param in kql_query.optional_params:
                    req_params[m_param] = ''
            missing_params = [p_name for p_name in missing_params
                              if p_name not in kql_query.optional_params]

        if missing_params:
            # If still have missing params then we error out
            query_help(kql_query.name)
            mssg = ('The following required parameters for this query were not set:',
                    ', '.join(missing_params))
            raise ValueError(mssg)

    return req_params


def _get_data_family_and_env(kql_query, providers, custom_params):
    """Get the data_family and environment."""
    data_family = None
    data_environment = None

    # If there is only one data family for this query, then use that
    if len(kql_query.data_families) == 1:
        data_family = kql_query.data_families[0]
    if len(kql_query.data_environments) == 1:
        data_environment = kql_query.data_environments[0]

    if data_family and data_environment:
        return data_family, data_environment

    candidate_families = set()
    candidate_environments = set()
    for provider in providers:
        if _DATA_FAMILY_NAME in provider.query_params:
            fam_value = DataFamily.parse(provider.query_params[_DATA_FAMILY_NAME])
            if fam_value:
                candidate_families.add(fam_value)
        if _DATA_ENVIRONMENT_NAME in provider.query_params:
            env_value = DataEnvironment.parse(provider.query_params[_DATA_ENVIRONMENT_NAME])
            if env_value:
                candidate_environments.add(env_value)

    if custom_params:
        # If we haven't yet worked out the data family and environment
        # try to get this from one of custom_params
        if _DATA_FAMILY_NAME in custom_params:
            fam_value = DataFamily.parse(custom_params[_DATA_FAMILY_NAME])
            if fam_value:
                candidate_families.add(fam_value)
        if _DATA_ENVIRONMENT_NAME in custom_params:
            env_value = DataEnvironment.parse(custom_params[_DATA_ENVIRONMENT_NAME])
            if env_value:
                candidate_environments.add(env_value)

    # get the intersection of families and environments that we found and those
    # supported by the query. If it is 1 item we are good to go.
    usable_families = candidate_families & set(kql_query.data_families)
    if len(usable_families) == 1:
        data_family = usable_families.pop()
    usable_environments = candidate_environments & set(kql_query.data_environments)
    if len(usable_environments) == 1:
        data_environment = usable_environments.pop()

    return data_family, data_environment


@export
def required_params(kql_query: any) -> list:
    """
    Return the set of required parameters for the query.

        :param query_string:
    """
    if isinstance(kql_query, KqlQuery):
        query_string = kql_query.query
    else:
        query_string = kql_query
    param_pattern = r'{([\w\d_-]+)}'
    return list(set(re.findall(param_pattern, query_string)))


def _add_queries_to_module(module_name):
    """Add queries to the module as callable methods."""
    if module_name not in sys.modules:
        raise LookupError(f'Module {module_name} was not found sys.modules')
    for query_name in query_definitions:
        module = sys.modules[module_name]
        query_func = partial(replace_prov_query_params, query_name=query_name)
        query_func.__doc__ = replace_prov_query_params.__doc__
        setattr(module, query_name, query_func)


# Add all queries defined in builtin queries module as functions
_add_queries_to_module(__name__)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""auditd extract test class."""
import ast
import unittest
import json
import os

import pandas as pd

from .. msticpy.sectools.auditdextract import extract_events_to_df

_test_data_folders = [d for d, _, _ in os.walk(os.getcwd()) if d.endswith('/tests/testdata')]
if len(_test_data_folders) == 1:
    _TEST_DATA = _test_data_folders[0]
else:
    _TEST_DATA = './tests/testdata'


class TestAuditdExtract(unittest.TestCase):
    """Unit test class."""

    def test_extract_from_log_analytics(self):
        input_file = os.path.join(_TEST_DATA, 'linux_events.csv')
        input_df = pd.read_csv(input_file)

        import_time = input_df['TimeGenerated'].iloc[0]
        self.assertGreaterEqual(input_df.shape[0], 100)

        input_df['AuditdMessage'] = input_df.apply(lambda x: ast.literal_eval(x.AuditdMessage), axis=1)

        output_df = extract_events_to_df(data=input_df)
        self.assertGreaterEqual(output_df.shape[0], 100, "Expected >=100 rows")
        self.assertGreaterEqual(output_df.shape[1], 32, "Expected >=32 colums")
        event_types = output_df['EventType'].drop_duplicates().to_list()
        self.assertIn('USER_END', event_types)
        self.assertIn('CRED_DISP', event_types)
        self.assertIn('SYSCALL_EXECVE', event_types)
        self.assertIn('USER_ACCT', event_types)
        self.assertIn('CRED_ACQ', event_types)
        self.assertIn('LOGIN', event_types)
        self.assertIn('USER_START', event_types)
        self.assertIn('USER_CMD', event_types)
        self.assertIn('CRED_REFR', event_types)

        self.assertGreaterEqual(len(output_df[output_df['EventType'] == 'SYSCALL_EXECVE']), 82, 'Expected EXECVE events')
        self.assertGreaterEqual(len(output_df[output_df['EventType'] == 'LOGIN']), 1, 'Expected LOGIN events')
        self.assertGreaterEqual(len(output_df[output_df['EventType'] == 'USER_CMD']), 2, 'Expected USER_CMD events')
        self.assertGreaterEqual(len(output_df[output_df['EventType'] == 'USER_ACCT']), 1, 'Expected USER_ACCT events')

        for _, execve_event in output_df[output_df['EventType'] == 'SYSCALL_EXECVE'][1:20].iterrows():
            self.assertIsNotNone(execve_event.Computer)
            self.assertIsNotNone(execve_event.TimeGenerated)
            self.assertNotEqual(import_time, execve_event.TimeGenerated)
            self.assertIsNotNone(execve_event.exe)
            self.assertIsNotNone(execve_event.cmdline)
            self.assertIsNotNone(execve_event.cwd)

        for _, login_event in output_df[output_df['EventType'] == 'LOGIN'].iterrows():
            self.assertIsNotNone(login_event.Computer)
            self.assertIsNotNone(login_event.acct)
            self.assertIsNotNone(login_event.TimeGenerated)
            self.assertNotEqual(import_time, login_event.TimeGenerated)


if __name__ == '__main__':
    unittest.main()
    print('bye')



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for pre-defined widget layouts."""

import json
from json import JSONDecodeError
import os
import re
from datetime import datetime, timedelta
from enum import Enum
from typing import Callable, List, Mapping

import pandas as pd
from IPython.display import display
import ipywidgets as widgets
from ipywidgets import Layout

from . import kql as qry
from .query_defns import QueryParamProvider
from .utility import export
from .. _version import VERSION

__version__ = VERSION
__author__ = 'Ian Hellen'


class TimeUnit(Enum):
    """Time unit enumeration and value."""

    sec = 1
    min = 60
    hour = 60 * 60
    day = 60 * 60 * 24


def _parse_time_unit(unit_str: str) -> TimeUnit:
    """Return the TimeUnit enum matching the input string."""
    if unit_str.startswith('m'):
        return TimeUnit.min
    if unit_str.startswith('h'):
        return TimeUnit.hour
    if unit_str.startswith('s'):
        return TimeUnit.sec
    if unit_str.startswith('d'):
        return TimeUnit.day
    return TimeUnit.min


@export
class Lookback(QueryParamProvider):
    """
    ipwidget wrapper to display integer slider.

    Parameters
    ----------
    QueryParamProvider : QueryParamProvider
        Abstract base class

    """

    def __init__(self, default: int = 4, label: str = 'Select time ({units}) to look back',
                 origin_time: datetime = None, min_value: int = 1, max_value: int = 240,
                 units: str = 'hour', auto_display: bool = False):
        """
        Create an instance of the lookback slider widget.

        Parameters
        ----------
        default : int, optional
            The default 'lookback' time (the default is 4)
        label : str, optional
            The description to display
            (the default is 'Select time ({units}) to look back')
        origin_time : datetime, optional
            The origin time (the default is `datetime.utcnow()`)
        min_value : int, optional
            Minimum value (the default is 1)
        max_value : int, optional
            Maximum value (the default is 240)
        units : str, optional
            Time unit (the default is 'hour')
            Permissable values are 'day', 'hour', 'minute', 'second'
            These can all be abbreviated down to initial characters
            ('d', 'm', etc.)
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)

        """
        # default to now
        self.origin_time = datetime.utcnow() if origin_time is None else origin_time

        self._time_unit = _parse_time_unit(units)
        if '{units}' in label:
            label = label.format(units=self._time_unit.name)
        self._lookback_wgt = widgets.IntSlider(value=default,
                                               min=min_value,
                                               max=max_value,
                                               step=1,
                                               description=label,
                                               layout=Layout(
                                                   width='60%', height='50px'),
                                               style={'description_width': 'initial'})

        self.end = self.origin_time
        self._time_unit = _parse_time_unit(units)
        self.start = self.end - timedelta(seconds=(self._time_unit.value *
                                          self._lookback_wgt.value))

        self._lookback_wgt.observe(self._time_range_change, names='value')

        if auto_display:
            self.display()

    def display(self):
        """Display the interactive widgets."""
        display(self._lookback_wgt)

    @property
    def lookback(self):
        """Return current widget lookback value."""
        return self._lookback_wgt.value

    @property
    def value(self):
        """Return current widget lookback value."""
        return self._lookback_wgt.value

    def _time_range_change(self, change):
        del change
        self.start = (self.origin_time +
                      timedelta(0, self._lookback_wgt.value * self._time_unit.value))

    @property
    def query_params(self):
        """
        Query parameters derived from alert.

        Returns:
            dict(str, str) -- Dictionary of parameter names

        """
        return {
            'start': self.start,
            'end': self.end
        }


@export
class QueryTime(QueryParamProvider):
    """
    QueryTime.

    Composite widget to capture date and time origin
    and set start and end times for queries.

    Parameters
    ----------
    QueryParamProvider : QueryParamProvider
        Abstract base class

    """

    _label_style = {'description_width': 'initial'}

    def __init__(self, origin_time: datetime = None, before: int = 60, after: int = 10,
                 max_before: int = 600, max_after: int = 100, label: str = None,
                 units: str = 'min', auto_display: bool = False):
        """
        Create new instance of QueryTime.

        Parameters
        ----------
        origin_time : datetime, optional
            The origin time (the default is `datetime.utcnow()`)
        label : str, optional
            The description to display
            (the default is 'Select time ({units}) to look back')
        before : int, optional
            The default number of `units` before the `origin_time`
            (the default is 60)
        after : int, optional
            The default number of `units` after the `origin_time`
            (the default is 10)
        max_before : int, optional
            The largest value for `before` (the default is 600)
        max_after : int, optional
            The largest value for `after` (the default is 100)
        units : str, optional
            Time unit (the default is 'min')
            Permissable values are 'day', 'hour', 'minute', 'second'
            These can all be abbreviated down to initial characters
            ('d', 'm', etc.)
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)

        """
        self._label = 'Set query time boundaries' if label is None else label
        self._time_unit = _parse_time_unit(units)

        max_before = abs(max_before)
        max_after = abs(max_after)
        before = abs(before)
        after = abs(after)
        if max_before < before:
            before = max_before
        if max_after < after:
            after = max_after

        # default to now
        self.origin_time = datetime.utcnow() if origin_time is None else origin_time
        # Calculate time offsets from origin
        self._query_start = self.origin_time - \
            timedelta(0, before * self._time_unit.value)
        self._query_end = self.origin_time + \
            timedelta(0, after * self._time_unit.value)

        # Create widgets
        self._w_origin_dt = widgets.DatePicker(description='Origin Date',
                                               disabled=False,
                                               value=self.origin_time.date())
        self._w_origin_tm = widgets.Text(description='Time (24hr)',
                                         disabled=False,
                                         value=str(self.origin_time.time()))

        range_desc = 'Time Range ({}):'.format(self._time_unit.name)
        self._w_tm_range = widgets.IntRangeSlider(value=[-before, after],
                                                  min=-max_before,
                                                  max=max_after,
                                                  step=1,
                                                  description=range_desc,
                                                  disabled=False,
                                                  continuous_update=True,
                                                  orientation='horizontal',
                                                  readout=True, readout_format='d',
                                                  layout=Layout(width='80%'),
                                                  style=self._label_style)

        self._w_start_time_txt = widgets.Text(value=self._query_start.isoformat(sep=' '),
                                              description='Query start time (UTC):',
                                              layout=Layout(width='50%'),
                                              style=self._label_style)
        self._w_end_time_txt = widgets.Text(value=self._query_end.isoformat(sep=' '),
                                            description='Query end time (UTC) :  ',
                                            layout=Layout(width='50%'),
                                            style=self._label_style)

        self._w_tm_range.observe(self._time_range_change, names='value')
        self._w_origin_dt.observe(self._update_origin, names='value')
        self._w_origin_tm.observe(self._update_origin, names='value')

        if auto_display:
            self.display()

    def display(self):
        """Display the interactive widgets."""
        display(widgets.HTML('<h4>{}</h4>'.format(self._label)))
        display(widgets.HBox([self._w_origin_dt, self._w_origin_tm]))
        display(widgets.VBox([self._w_tm_range,
                              self._w_start_time_txt,
                              self._w_end_time_txt]))

    # pylint: disable=locally-disabled, W0613
    def _update_origin(self, change):
        try:
            tm_value = datetime.strptime(
                self._w_origin_tm.value, '%H:%M:%S.%f').time()
            self.origin_time = datetime.combine(
                self._w_origin_dt.value, tm_value)
            self._time_range_change(change=None)
        except ValueError:
            pass

    def _time_range_change(self, change):
        self._query_start = (self.origin_time +
                             timedelta(0, self._w_tm_range.value[0] * self._time_unit.value))
        self._query_end = (self.origin_time +
                           timedelta(0, self._w_tm_range.value[1] * self._time_unit.value))
        self._w_start_time_txt.value = self._query_start.isoformat(sep=' ')
        self._w_end_time_txt.value = self._query_end.isoformat(sep=' ')

    @property
    def start(self):
        """Query start time."""
        return self._query_start

    @property
    def end(self):
        """Query end time."""
        return self._query_end

    @property
    def query_params(self):
        """
        Query parameters derived from alert.

        Returns:
            dict(str, str) -- Dictionary of parameter names

        """
        return {
            'start': self.start,
            'end': self.end
        }


@export
class AlertSelector(QueryParamProvider):
    """
    AlertSelector.

    View list of alerts and select one for investigation.
    Optionally provide and action to call with the selected alert as a parameter
    (typically used to display the alert.)

    Attributes:
        selected_alert: the selected alert
        alert_id: the ID of the selected alert
        alerts: the current alert list (DataFrame)

    """

    _ALERTID_REGEX = r'\[id:(?P<alert_id>.*)\]$'

    def __init__(self, alerts: pd.DataFrame, action: Callable[..., None] = None,
                 columns: list({str})=None, auto_display: bool = False):
        """
        Create a new instance of AlertSelector.

        Parameters
        ----------
        alerts : pd.DataFrame
            DataFrame of alerts.
        action : Callable[..., None], optional
            Optional function to execute for each selected alert.
            (the default is None)
        columns : list, optional
            Override the default column names to use from `alerts`
            (the default is ['StartTimeUtc', 'AlertName',
            'CompromisedEntity', 'SystemAlertId'])
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)

        """
        self.alerts = alerts
        self.alert_action = action

        if not columns:
            columns = ['StartTimeUtc', 'AlertName',
                       'CompromisedEntity', 'SystemAlertId']
        items = alerts[columns]
        items = items.sort_values('StartTimeUtc', ascending=True)
        self._select_items = \
            items.apply(self._alert_summary, axis=1).values.tolist()

        self.selected_alert = None
        self.alert_id = None

        self._w_select_alert = widgets.Select(options=self._select_items,
                                              description='Select alert :',
                                              layout=Layout(
                                                  width='95%', height='300px'),
                                              style={'description_width': 'initial'})

        self._w_filter_alerts = widgets.Text(value='', description='Filter alerts by title:',
                                             style={'description_width': 'initial'})
        self._w_output = widgets.Output(layout={'border': '1px solid black'})

        # set up observer callbacks
        self._w_filter_alerts.observe(self._update_options, names='value')
        self._w_select_alert.observe(self._select_alert, names='value')

        if auto_display:
            self.display()

    def display(self):
        """Display the interactive widgets."""
        self._select_top_alert()
        display(widgets.VBox([self._w_filter_alerts,
                              self._w_select_alert,
                              self._w_output]))

    def _alert_summary(self, alert_row):
        """Return summarized string of alert properties."""
        return '{time}  {alert} ({host}) [id:{id}]'.format(time=alert_row.StartTimeUtc,
                                                           alert=alert_row.AlertName,
                                                           host=alert_row.CompromisedEntity,
                                                           id=alert_row.SystemAlertId)

    def _update_options(self, change):
        """Filter the alert list by substring."""
        if change is not None and 'new' in change:
            self._w_select_alert.options = [
                i for i in self._select_items if change['new'].lower() in i.lower()]

    def _select_alert(self, selection=None):
        """Select action triggered by picking item from list."""
        if (selection is None or 'new' not in selection or
                not isinstance(selection['new'], str)):
            self.selected_alert = None
        else:
            match = re.search(self._ALERTID_REGEX, selection['new'])
            if match is not None:
                self.alert_id = match.groupdict()['alert_id']
                self.selected_alert = self._get_alert(self.alert_id)
                if self.alert_action is not None:
                    self._w_output.clear_output()
                    with self._w_output:
                        self.alert_action(self.selected_alert)

    def _get_alert(self, alert_id):
        """Get the alert by alert_id."""
        self.alert_id = alert_id
        selected_alerts = self.alerts[self.alerts['SystemAlertId'] == alert_id]

        if selected_alerts.shape[0] > 0:
            alert = pd.Series(selected_alerts.iloc[0])
            if isinstance(alert['ExtendedProperties'], str):
                try:
                    alert['ExtendedProperties'] = json.loads(
                        (alert['ExtendedProperties']))
                except JSONDecodeError:
                    pass
            if isinstance(alert['Entities'], str):
                try:
                    alert['Entities'] = json.loads((alert['Entities']))
                except JSONDecodeError:
                    pass
            return alert

    def _select_top_alert(self):
        """Select the first alert by default."""
        top_alert = self.alerts.iloc[0]
        if len(top_alert) == 1:
            self.alert_id = top_alert.SystemAlertId
            self.selected_alert = self._get_alert(self.alert_id)
            if self.alert_action is not None:
                self._w_output.clear_output()
                with self._w_output:
                    self.alert_action(self.selected_alert)

    @property
    def query_params(self):
        """
        Query parameters derived from alert.

        Returns:
            dict(str, str) -- Dictionary of parameter names

        """
        return {
            'provider_alert_id': self.alert_id
        }


@export
class GetSingleAlert(QueryParamProvider):
    """
    GetSingleAlert.

    Try to fetch a single alert by SystemAlertId.

    Attributes:
        selected_alert: the selected alert
        alert_id: the ID of the selected alert
        alerts: the current alert list (DataFrame)

    """

    def __init__(self, action: Callable[..., None] = None, max_lookback: float = 28,
                 query_time_provider=None, auto_display: bool = False):
        """
        Create a new instance of GetSingleAlert.

        Parameters
        ----------
        action : Callable[..., None], optional
            Create a new instance of GetSingleAlert.
            (the default is None)
        max_lookback : float, optional
            Number of days to search for an alert with the
            supplied ID (the default is 28, fractional days allowed)
        query_time_provider : [type], optional
            An object with 'start' and 'end' properties
            or a QueryParamProvider with start and end defined in its
            query_params property (the default is None)
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)

        """
        self.alert_action = action
        self.selected_alert = None
        self.alert_id = None
        self.alerts = None

        self._start = None
        self._end = None
        if query_time_provider is not None:
            if ('start' in query_time_provider.__dir__ and
                    'end' in query_time_provider.__dir__):
                self._start = query_time_provider.start
                self._end = query_time_provider.end
            elif isinstance(query_time_provider, QueryParamProvider):
                self._start = query_time_provider.query_params.get('start', None)
                self._end = query_time_provider.query_params.get('end', None)

        if self._end is None:
            self._end = datetime.now()
            self._start = self._end - timedelta(max_lookback)

        self._w_target_alert = widgets.Text(
            value=self.alert_id,
            placeholder='SystemAlertId',
            description='SystemAlertId for alert :',
            layout=Layout(width='50%'),
            style={'description_width': 'initial'})

        self._w_fetch_button = widgets.Button(description="Get alert..")
        self._w_fetch_button.on_click(self._click_get_alert)

        self._w_output = widgets.Output(layout={'border': '1px solid black'})

        if auto_display:
            self.display()

    @property
    def query_params(self):
        """
        Query parameters derived from alert.

        Returns:
            dict(str, str) -- Dictionary of parameter names

        """
        return {
            'system_alert_id': self.alert_id
        }

    def display(self):
        """Display the interactive widgets."""
        display(widgets.VBox([self._w_target_alert,
                              self._w_fetch_button,
                              self._w_output]))

    def _click_get_alert(self, button):
        del button
        self.alert_id = self._w_target_alert.value
        if not self.alert_id or not self.alert_id.strip():
            print('Error: AlertID was not entered')

        self.alerts = qry.exec_query(query_name='get_alert',
                                     start=self._start,
                                     end=self._end,
                                     system_alert_id=self.alert_id)
        if self.alerts is not None:
            self.selected_alert = self._get_alert(self.alert_id)
            if self.alert_action is not None:
                self._w_output.clear_output()
                with self._w_output:
                    self.alert_action(self.selected_alert)
        else:
            print('Alert not found.')

    def _get_alert(self, alert_id):
        self.alert_id = alert_id
        selected_alerts = self.alerts[self.alerts['SystemAlertId'] == alert_id]

        if selected_alerts.shape[0] > 0:
            alert = pd.Series(selected_alerts.iloc[0])
            if isinstance(alert['ExtendedProperties'], str):
                alert['ExtendedProperties'] = json.loads(
                    (alert['ExtendedProperties']))
            if isinstance(alert['Entities'], str):
                try:
                    alert['Entities'] = json.loads((alert['Entities']))
                except JSONDecodeError:
                    pass
            return alert


@export
class GetEnvironmentKey:
    """
    GetEnvironmentKey.

    Tries to retrieve an environment variable value. The value
    can be changed/set and optionally saved back to the system
    environment.
    """

    def __init__(self, env_var: str, help_str: str = None,
                 prompt: str = "Enter the value: ",
                 auto_display: bool = False):
        """
        Create a new instance of GetEnvironmentKey.

        Parameters
        ----------
        env_var : str
            Name of the environment variable.
        help_str : str, optional
            Help to display if the environment variable is not set. (the default is None)
        prompt : str, optional
            Prompt to display with the text box.
            (the default is "Enter the value: ")
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)

        """
        self._value = os.environ.get(env_var)
        self._name = env_var

        if not self._value:
            display(widgets.HTML(value=help_str))

        self._w_text = widgets.Text(value=self._value,
                                    description=prompt, layout=Layout(
                                        width='50%'),
                                    style={'description_width': 'initial'})

        self._w_save_button = widgets.Button(description='Set',
                                             layout=Layout(
                                                 width='10%', disabled=False),
                                             style={'description_width': 'initial'})
        self._w_check_save = widgets.Checkbox(value=True,
                                              description='Save as environment var',
                                              disabled=False)
        self._w_save_button.on_click(self._on_save_button_clicked)
        self._hbox = widgets.HBox(
            [self._w_text, self._w_save_button, self._w_check_save])

        if auto_display:
            self.display()

    @property
    def value(self):
        """Get the current value of the key."""
        return self._w_text.value.strip()

    @property
    def name(self):
        """Get the current name of the key."""
        return self._name

    def display(self):
        """Display the interactive widgets."""
        display(self._hbox)

    # pylint: disable=locally-disabled, W0613
    def _on_save_button_clicked(self, button):
        if self._w_check_save.value:
            os.environ[self._name] = self._w_text.value.strip()


@export
class SelectString:
    """
    Selection list from list or dict.

    Attributes:
        value : The selected value.

    """

    def __init__(self, 
                 description: str = 'Select an item',
                 item_list: List[str] = None,
                 action: Callable[..., None] = None,
                 item_dict: Mapping[str, str] = None,
                 auto_display: bool = False,
                 height: str = '100px',
                 width: str = '50%'):
        """
        Select an item from a list or dict.

        Parameters
        ----------
        description : str, optional
            The widget label to display.
            (the default is 'Select an item')
        item_list : List[str], optional
            A `list` of items to select from (the default is None)
        item_dict : Mapping[str, str], optional
            A `dict` of items to select from. When using `item_dict`
            the keys are displayed as the selectable items and value
            corresponding to the selected key is set as the `value`
            property.
            (the default is None)
        action : Callable[..., None], optional
            function to call when item selected (passed a single
            parameter - the value of the currently selected item)
            (the default is None)
        auto_display : bool, optional
            Whether to display on instantiation (the default is False)
        height : str, optional
            Selection list height (the default is '100px')
        width : str, optional
            Selection list width (the default is '50%')

        """
        if item_list:
            self._item_list = item_list
            self._item_dict = None
            self.value = item_list[0]
        elif item_dict:
            self._item_list = list(item_dict.keys())
            self._item_dict = item_dict
            self.value = list(self._item_dict.values())[0]
        else:
            raise ValueError(
                "One of item_list or item_dict must be supplied.")

        self.item_action = action
        self._wgt_select = widgets.Select(options=self._item_list,
                                          description=description,
                                          layout=Layout(
                                              width=width, height=height),
                                          style={'description_width': 'initial'})
        self._wgt_select.observe(self._select_item, names='value')

        if auto_display:
            self.display()

    def _select_item(self, selection):
        if (selection is None or 'new' not in selection or
                not isinstance(selection['new'], str)):
            return
        value = selection['new']

        if self._item_dict:
            self.value = self._item_dict.get(value, None)
        else:
            self.value = value

        if self.item_action is not None:
            self.item_action(self.value)

    def display(self):
        """Display the interactive widget."""
        display(self._wgt_select)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Jupyter Notebook Security Tools."""
# flake8: noqa: F403

# pylint: disable=W0401
from . import query_builtin_queries
from . nbwidgets import *
from . entityschema import *
from . import kql as qry
from . security_alert import *
from . security_event import *
from . security_alert_graph import *
from . import utility as util
from . query_mgr import *
from . import query_builtin_queries as qrydef
from . import nbdisplay as disp

# pylint: enable=W0401



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""MSTIC Security Tools."""

# flake8: noqa: F403
from . import base64unpack as b64
from . iocextract import *
from . vtlookup import *
from . eventcluster import *



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for Log Analytics-related configuration."""

import json
from .utility import export

__version__ = '0.1'
__author__ = 'Ian Hellen'

@export
class WorkspaceConfig():
    """Workspace configuration class."""

    # Constants
    TENANT_ID = "{{cookiecutter.tenant_id}}"
    SUBSCRIPTION_ID = "{{cookiecutter.subscription_id}}"
    RESOURCE_GROUP = "{{cookiecutter.resource_group}}"
    WORKSPACE_ID = "{{cookiecutter.workspace_id}}"
    WORKSPACE_NAME = "{{cookiecutter.workspace_name}}"

    def __init__(self, config_file: str):
        """
        Load current Azure Notebooks configuration for Log Analytics.

        Arguments:
            config_file {str} -- path to the configuration file.
        """
        self._config_file = config_file
        self._config = self._read_config_values(config_file)

    def __getitem__(self, key: str):
        """Allow property get using dictionary key syntax."""
        if key in self._config:
            return self._config[key]
        raise KeyError

    def __setitem__(self, key: str, value: any):
        """Allow property set using dictionary key syntax."""
        self._config[key] = value

    def __contains__(self, key: str):
        """Allow property in test."""
        # In operator overload
        return (key == 'Type' or
                key in self._config or
                key in self.__dict__)

    @classmethod
    def _read_config_values(cls, file_path: str) -> dict:
        """Read configuration file."""
        with open(file_path) as json_file:
            if json_file:
                json_config = json.load(json_file)
                return json_config



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for IoCExtract class."""

import re
from collections import namedtuple, defaultdict
from typing import Any, List, Mapping, Set
from urllib.parse import unquote

import pandas as pd
from .. nbtools.utility import export
from .. _version import VERSION

__version__ = VERSION
__author__ = 'Ian Hellen'


def _compile_regex(regex):
    return re.compile(regex, re.I | re.X | re.M)


IoCPattern = namedtuple('IoCPattern', ['ioc_type', 'comp_regex', 'priority'])


@export
class IoCExtract(object):
    """
    IoC Extractor - looks for common IoC patterns in input strings.

    The extract() method takes either a string or a pandas DataFrame
    as input. When using the string option as an input extract will
    return a dictionary of results. When using a DataFrame the results
    will be returned as a new DataFrame with the following columns:
    IoCType: the mnemonic used to distinguish different IoC Types
    Observable: the actual value of the observable
    SourceIndex: the index of the row in the input DataFrame from
    which the source for the IoC observable was extracted.

    The class has a number of built-in IoC regex definitions.
    These can be retrieved using the ioc_types attribute.

    Addition IoC definitions can be added using the add_ioc_type
    method.

    Note: due to some ambiguity in the regular expression patterns
    for different types and observable may be returned assigned to
    multiple observable types. E.g. 192.168.0.1 is a also a legal file
    name in both Linux and Windows. Linux file names have a particularly
    large scope in terms of legal characters so it will be quite common
    to see other IoC observables (or parts of them) returned as a
    possible linux path.
    """

    IPV4_REGEX = r'(?P<ipaddress>(?:[0-9]{1,3}\.){3}[0-9]{1,3})'
    IPV6_REGEX = r'(?<![:.\w])(?:[A-F0-9]{1,4}:){7}[A-F0-9]{1,4}(?![:.\w])'
    DNS_REGEX = r'((?=[a-z0-9-]{1,63}\.)[a-z0-9]+(-[a-z0-9]+)*\.){2,}[a-z]{2,63}'
    # dns_regex = '\\b((?=[a-z0-9-]{1,63}\\.)[a-z0-9]+(-[a-z0-9]+)*\\.){2,}[a-z]{2,63}\\b'

    URL_REGEX = r'''
            (?P<protocol>(https?|ftp|telnet|ldap|file)://)
            (?P<userinfo>([a-z0-9-._~!$&\'()*+,;=:]|%[0-9A-F]{2})*@)?
            (?P<host>([a-z0-9-._~!$&\'()*+,;=]|%[0-9A-F]{2})*)
            (:(?P<port>\d*))?
            (/(?P<path>([^?\# ]|%[0-9A-F]{2})*/?))?
            (\?(?P<query>([a-z0-9-._~!$&'()*+,;=:/?@]|%[0-9A-F]{2})*))?
            (\#(?P<fragment>([a-z0-9-._~!$&'()*+,;=:/?@]|%[0-9A-F]{2})*))?'''

    WINPATH_REGEX = r'''
            (?P<root>[a-z]:|\\\\[a-z0-9_.$-]+||[.]+)
            (?P<folder>\\(?:[^\/:*?"\'<>|\r\n]+\\)*)
            (?P<file>[^\\/*?""<>|\r\n ]+)'''
    # Linux simplified - this ignores some legal linux paths avoid matching too much
    # TODO - also matches URLs!
    LXPATH_REGEX = r'''(?P<root>/+||[.]+)
            (?P<folder>/(?:[^\\/:*?<>|\r\n]+/)*)
            (?P<file>[^/\0<>|\r\n ]+)'''

    MD5_REGEX = r'(?:^|[^A-Fa-f0-9])(?P<hash>[A-Fa-f0-9]{32})(?:$|[^A-Fa-f0-9])'
    SHA1_REGEX = r'(?:^|[^A-Fa-f0-9])(?P<hash>[A-Fa-f0-9]{40})(?:$|[^A-Fa-f0-9])'
    SHA256_REGEX = r'(?:^|[^A-Fa-f0-9])(?P<hash>[A-Fa-f0-9]{64})(?:$|[^A-Fa-f0-9])'

    _content_regex = {}

    def __init__(self):
        """Intialize new instance of IoCExtract."""
        # IP Addresses
        self.add_ioc_type('ipv4', self.IPV4_REGEX, 0)
        self.add_ioc_type('ipv6', self.IPV6_REGEX, 0)

        # Dns Domains
        # TODO - This also matches IP addresses
        self.add_ioc_type('dns', self.DNS_REGEX, 1)

        # Http requests
        self.add_ioc_type('url', self.URL_REGEX, 0)

        # File paths
        # Windows
        self.add_ioc_type('windows_path', self.WINPATH_REGEX, 2)

        self.add_ioc_type('linux_path', self.LXPATH_REGEX, 2)

        # MD5, SHA1, SHA256 hashes
        self.add_ioc_type('md5_hash', self.MD5_REGEX, 1)
        self.add_ioc_type('sha1_hash', self.SHA1_REGEX, 1)
        self.add_ioc_type('sha256_hash', self.SHA256_REGEX, 1)

    # Public members

    def add_ioc_type(self, ioc_type: str, ioc_regex: str, priority: int = 0):
        """
        Add an IoC type and regular expression to use to the built-in set.

        Parameters
        ----------
        ioc_type : str
            A unique name for the IoC type
        ioc_regex : str
            A regular expression used to search for the type
        priority : int, optional
            Priority of the regex match vs. other ioc_patterns. 0 is
            the highest priority (the default is 0).

        Notes
        -----
        Pattern priorities.
            If two IocType patterns match on the same substring, the matched
            substring is assigned to the pattern/IocType with the highest
            priority. E.g. `foo.bar.com` will match types: `dns`, `windows_path`
            and `linux_path` but since `dns` has a higher priority, the expression
            is assigned to the `dns` matches.

        """
        if ioc_type is None or ioc_type.strip() is None:
            raise Exception('No value supplied for ioc_type parameter')
        if ioc_regex is None or ioc_regex.strip() is None:
            raise Exception('No value supplied for ioc_regex parameter')

        self._content_regex[ioc_type] = IoCPattern(ioc_type=ioc_type,
                                                   comp_regex=_compile_regex(
                                                       regex=ioc_regex),
                                                   priority=priority)

    @property
    def ioc_types(self) -> dict:
        """
        Return the current set of IoC types and regular expressions.

        Returns
        -------
        dict
            dict of IoC Type names and regular expressions

        """
        return self._content_regex

    def extract(self, src: str = None,
                data: pd.DataFrame = None,
                columns: List[str] = None,
                os_family='Windows',
                ioc_types: List[str] = None,
                include_paths: bool = False) -> Any:
        """
        Extract IoCs from either a string or pandas DataFrame.

        Parameters
        ----------
        src : str, optional
            source string in which to look for IoC patterns
            (the default is None)
        data : pd.DataFrame, optional
            input DataFrame from which to read source strings
            (the default is None)
        columns : list, optional
            The list of columns to use as source strings,
            if the `data` parameter is used. (the default is None)
        os_family : str, optional
            'Linux' or 'Windows' (the default is 'Windows'). This
            is used to toggle between Windows or Linux path matching.
        ioc_types : list, optional
            Restrict matching to just specified types.
            (default is all types)
        include_paths : bool, optional
            Whether to include path matches (which can be noisy)
            (the default is false - excludes 'windows_path'
            and 'linux_path'). If `ioc_types` is specified
            this parameter is ignored.

        Returns
        -------
        Any
            dict of found observables (if input is a string) or
            DataFrame of observables

        Notes
        -----
        Extract takes either a string or a pandas DataFrame as input.
        When using the string option as an input extract will
        return a dictionary of results.
        When using a DataFrame the results will be returned as a new
        DataFrame with the following columns:
        - IoCType: the mnemonic used to distinguish different IoC Types
        - Observable: the actual value of the observable
        - SourceIndex: the index of the row in the input DataFrame from
        which the source for the IoC observable was extracted.

        IoCType Pattern selection
        The default list is:  ['ipv4', 'ipv6', 'dns', 'url',
        'md5_hash', 'sha1_hash', 'sha256_hash'] plus any
        user-defined types.
        'windows_path', 'linux_path' are excluded unless `include_paths`
        is True or explicitly included in `ioc_paths`.

        """
        if src and src.strip():
            return self._scan_for_iocs(src, os_family)

        if data is None:
            raise Exception('No source data was supplied to extract')

        # Handle DataFrame option
        assert isinstance(data, pd.DataFrame)

        if columns is None:
            raise Exception(
                'No values where supplied for the columns parameter')

        # Use only requested IoC Type patterns
        if ioc_types:
            ioc_types_to_use = list(set(ioc_types))
        else:
            ioc_types_to_use = list(set(self._content_regex.keys()))
            if not include_paths:
                ioc_types_to_use.remove('windows_path')
                ioc_types_to_use.remove('linux_path')

        col_set = set(columns)
        if not col_set <= set(data.columns):
            missing_cols = [
                elem for elem in col_set if elem not in data.colums]
            raise Exception('Source column(s) {} not found in supplied DataFrame'
                            .format(', '.join(missing_cols)))

        result_columns = ['IoCType', 'Observable', 'SourceIndex']
        result_frame = pd.DataFrame(columns=result_columns)
        for idx, datarow in data.iterrows():
            for col in columns:
                ioc_results = self._scan_for_iocs(
                    datarow[col], os_family, ioc_types_to_use)
                for result_type, result_set in ioc_results.items():
                    if result_set:
                        for observable in result_set:
                            result_row = pd.Series(
                                data=[result_type, observable, idx], index=result_columns)
                            result_frame = result_frame.append(
                                result_row, ignore_index=True)

        return result_frame

    def validate(self, input_str: str, ioc_type: str) -> bool:
        """
        Check that `input_str` matches the regex for the specificed `ioc_type`.

        Parameters
        ----------
        input_str : str
            the string to test
        ioc_type : str
            the regex pattern to use

        Returns
        -------
        bool
            True if match.

        """
        if ioc_type not in self._content_regex:
            raise KeyError('Unknown type {}. Valid types are: {}'
                           .format(ioc_type, list(self._content_regex.keys())))
        rgx = self._content_regex[ioc_type]
        return rgx.comp_regex.fullmatch(input_str) is not None

    # Private methods
    def _scan_for_iocs(self, src: str,
                       os_family: str,
                       ioc_types: List[str] = None) -> Mapping[str, Set[str]]:
        """Return IoCs found in the string."""
        ioc_results = defaultdict(set)
        iocs_found = {}

        for (ioc_type, rgx_def) in self._content_regex.items():
            if ioc_types and ioc_type not in ioc_types:
                continue

            if os_family == 'Linux' and rgx_def.ioc_type == 'windows_path':
                continue
            elif os_family == 'Windows' and rgx_def.ioc_type == 'linux_path':
                continue

            match_pos = 0
            for rgx_match in rgx_def.comp_regex.finditer(src, match_pos):
                if rgx_match is not None:
                    self._add_highest_pri_match(iocs_found,
                                                rgx_match.group(),
                                                rgx_def)
                    if ioc_type == 'url':
                        decoded_url = unquote(rgx_match.group())
                        for url_match in rgx_def.comp_regex.finditer(decoded_url, match_pos):
                            if url_match is not None:
                                self._add_highest_pri_match(iocs_found,
                                                            url_match.group(),
                                                            rgx_def)
                                self._add_highest_pri_match(iocs_found,
                                                            url_match.groupdict()[
                                                                'host'],
                                                            self._content_regex['dns'])
                    match_pos = rgx_match.end()
                else:
                    break
        for ioc, ioc_result in iocs_found.items():
            ioc_results[ioc_result[0]].add(ioc)

        return ioc_results

    def _add_highest_pri_match(self,
                               iocs_found: dict,
                               current_match: str,
                               current_def: IoCPattern):
        # if we already found a match for this item and the previous
        # ioc type is more specific then don't add this to the results
        if current_match in iocs_found and current_def.priority > iocs_found[current_match][1]:
            return

        iocs_found[current_match] = (
            current_def.ioc_type, current_def.priority)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for VTLookup class."""

import json
from json import JSONDecodeError
import math
import re
from typing import List, Mapping, Any
from collections import namedtuple, Counter
from ipaddress import IPv4Address, ip_address
import socket
from socket import gaierror

import pandas as pd
import requests
from urllib3.exceptions import LocationParseError
from urllib3.util import parse_url

from . iocextract import IoCExtract
from .. nbtools.utility import export, pd_version_23
from .. _version import VERSION

__version__ = VERSION
__author__ = 'Ian Hellen'

# VirusTotal parameter collection
VTParams = namedtuple('VTParams',
                      ['api_type', 'batch_size', 'batch_delimiter',
                       'http_verb', 'api_var_name', 'headers'])

DuplicateStatus = namedtuple('DuplicateStatus', ['is_dup', 'status'])
PreProcessResult = namedtuple('PreProcessResult', ['observable', 'status'])


@export
class VTLookup:
    """
    VTLookup: VirusTotal lookup of IoC reports.

    Main methods are:
    lookup_iocs() - accepts input of multiple IoCs in a Pandas DataFrame
    lookup_ioc() - looks up a single IoC observable.
    supported_ioc_types - a list of valid target types.
    ioc_vt_type_mapping - a dictionary of mappings to recognized VT Types.
    Types mapped to None will not be submitted to VT.

    For urls a full http request can be submitted, query string and fragments will be
    dropped before submitting.
    For files MD5, SHA1 and SHA256 hashes are supported.
    For IP addresses only dotted IPv4 addresses are supported.
    """

    # Ioc types that we support
    _SUPPORTED_INPUT_TYPES = ['ipv4', 'dns', 'url', 'md5_hash',
                              'sha1_hash', 'sh256_hash']

    # Mapping to to VT Types
    _VT_TYPE_MAP = {'ipv4': 'ip-address',
                    'ipv6': None,
                    'dns': 'domain',
                    'url': 'url',
                    'md5_hash': 'file',
                    'sha1_hash': 'file',
                    'sh256_hash': 'file'}

    # VT API parameters
    _HDR_GZIP = {"Accept-Encoding": "gzip, deflate"}
    _VT_API = 'https://www.virustotal.com/vtapi/v2/{type}/report'
    _VT_API_TYPES = {'url': VTParams('url', 1, '\n', 'get', 'resource', _HDR_GZIP),
                     'file': VTParams('file', 25, ',', 'get', 'resource', _HDR_GZIP),
                     'ip-address': VTParams('ip-address', 1, '', 'get', 'ip', None),
                     'domain': VTParams('domain', 1, '', 'get', 'domain', None)}

    _RESULT_COLUMNS = ['Observable', 'IoCType', 'Status', 'ResponseCode', 'RawResponse',
                       'Resource', 'SourceIndex', 'VerboseMsg', 'Resource', 'ScanId',
                       'Permalink', 'Positives', 'MD5', 'SHA1', 'SHA256',
                       'ResolvedDomains', 'ResolvedIPs', 'DetectedUrls']

    def __init__(self, vtkey: str, verbosity: int = 1):
        """
        Create a new instance of VTLookup class.

        Parameters
        ----------
        vtkey : str
            VirusTotal API key
        verbosity : int, optional
            The level of detail of reporting
                0 = no reporting
                1 = minimal reporting (default)
                2 = verbose reporting

        """
        self._vtkey = vtkey
        self._verbosity = verbosity
        self._ioc_custom_type_map = {}

        self._ioc_extract = IoCExtract()

        # create a data frame to store the results
        self.results = pd.DataFrame(data=None, columns=self._RESULT_COLUMNS)

    @property
    def supported_ioc_types(self) -> List[str]:
        """
        Return list of supported IoC type internal names.

        Returns
        -------
        List[str]
            List of supported IoC type internal names.

        """
        return self._SUPPORTED_INPUT_TYPES

    @property
    def supported_vt_types(self) -> List[str]:
        """
        Return list of VirusTotal supported IoC type names.

        Returns
        -------
        List[str]
            List of VirusTotal supported IoC type names.

        """
        return list(self._VT_API_TYPES.keys())

# flake8: noqa: D102
    @property
    def ioc_vt_type_mapping(self) -> Mapping[str, str]:
        """
        Return mapping between internal and VirusTotal IoC type names.

        Returns
        -------
        Mapping[str, str]
            Return mapping between internal and VirusTotal IoC type names.

        """
        return self._VT_TYPE_MAP

    def lookup_iocs(self, data: pd.DataFrame, src_col: str = 'Observable',
                    type_col: str = 'IoCType', src_index_col: str = 'SourceIndex',
                    **kwargs) -> pd.DataFrame:
        """
        Retrieve results for IoC observables in the source dataframe.

        Parameters
        ----------
        data : pd.DataFrame
            Dataframe containing the observables to search for
        src_col : str, optional
            The column name that contains the observable data
            (one item per row) (the default is 'Observable')
        type_col : str, optional
            The column name containing the observable type
            (the default is 'IoCType')
        src_index_col : str, optional
            The name of the column to use as source index. If not
            specified this defaults to 'SourceIndex'. If this (or the supplied value)
            is not in the source dataframe the index of the source dataframe will
            be used. This is retained in the output so that you can join the results
            back to the original data.
            (the default is 'SourceIndex')

        Other Parameters
        ----------------
            key/value pairs of additional mappings to supported IoC type names
            e.g. ipv4='ipaddress', url='httprequest'. This allows you to specify custom
            mappings when the source data is tagged with different names.

        Returns
        -------
        pd.DataFrame
            Combined results of local pre-processing and VirusTotal Lookups

        Notes
        -----
            See supported_ioc_types attribute for a list of valid target types.
            Not all of these types are supported by VirusTotal. See ioc_vt_type_mapping for
            current mappings. Types mapped to None will not be submitted to VT.

            For urls a full http request can be submitted, query string and fragments will be
            dropped before submitting. Other supported protocols are ftp, telnet, ldap, file
            For files MD5, SHA1 and SHA256 hashes are supported.
            For IP addresses only dotted IPv4 addresses are supported.

        """
        # if the caller has supplied alternative type name mappings add any of these
        # to our lookup dictionary
        for k in self._get_supported_vt_ioc_types():
            self._ioc_custom_type_map[k] = k
        for k, val in kwargs.items():
            if k in self._get_supported_vt_ioc_types():
                self._ioc_custom_type_map[k] = val

        if src_index_col not in data:
            src_index_col = None

        # for each ioc_type, retrieve observables from dataframe
        for ioc_type, mapped_type in self._ioc_custom_type_map.items():
            input_df = data[data[type_col] == mapped_type]
            self._lookup_ioc_type(input_df, ioc_type, src_col, src_index_col)

        self._print_status(
            'Submission complete. {} responses from {} input rows'
            .format(len(self.results), len(data)), 2)

        return self.results

    def lookup_ioc(self, observable: str,
                   ioc_type: str,
                   output: str = 'dict') -> Any:
        """
        Look up and single IoC observable.

        Parameters
        ----------
        observable : str
            The observable value
        ioc_type : str
            The IoC Type (see 'supported_ioc_types' attribute)
        output : str, optional
            Output results as a dictionary (or list of dicts)
            if `output` is any other value the result will be returned in a
            Pandas DataFrame (the default is 'dict')

        Returns
        -------
            list{dict}: if output == 'dict'
            pd.DataFrame: otherwise

        """
        # Check input
        if (observable is None or observable.strip() is None or
                ioc_type is None or ioc_type.strip() is None):
            raise SyntaxError("Invalid value for observable or ioc_type")

        observable, status = self._preprocess_observable(observable, ioc_type)
        if observable is None:
            raise SyntaxError(
                '{} for observable value {}'.format(status, observable), 1)

        if ioc_type not in self._VT_TYPE_MAP:
            raise LookupError('IoC Type {} not recognized. Valid types are [{}]'.format(
                ioc_type, ', '.join(self.supported_ioc_types)))

        if self._VT_TYPE_MAP[ioc_type] not in self._VT_API_TYPES:
            vt_types = {
                k for k, val in self.ioc_vt_type_mapping if val is not None}
            err = 'IoC Type {} is recognized by VirusTotal. Valid types are [{}]'.format(
                ioc_type, ', '.join(vt_types))
            raise LookupError(err)

        # do the submission
        vt_param = self._VT_API_TYPES[self._VT_TYPE_MAP[ioc_type]]
        results, _ = self._vt_submit_request(observable, vt_param)
        self._parse_vt_results(results, observable, ioc_type)

        # return as a list of dictionaries or a DataFrame
        if output == 'dict':
            list_res = self.results.apply(
                lambda x: x.to_dict(), axis=1).tolist()
            return list_res[0] if len(list_res) == 1 else list_res
        
        return self.results

    def _lookup_ioc_type(self, input_frame: pd.DataFrame, ioc_type: str, src_col: str,
                         src_index_col: str):
        """
        Perform the VT submission of a set of IoCs of a given type.

        Parameters
        ----------
        input_frame : pd.DataFrame
            the input dataframe
        ioc_type : str
            the IoC Type to submit
        src_col : str
            The name column in the dataframe containing the
            IoC observables
        src_index_col : str
            SourceIndex column name

        """
        assert(ioc_type in self._VT_TYPE_MAP and
               self._VT_TYPE_MAP[ioc_type] in self._VT_API_TYPES)
        vt_param = self._VT_API_TYPES[self._VT_TYPE_MAP[ioc_type]]

        # Some types support batch lookups so we can assemble them into batches
        # for the moment we are only supporting
        source_row_index = {}
        obs_batch = []
        batch_index = 0
        row_num = 0
        row_count = len(input_frame)
        if src_index_col:
            src_cols = [src_col, src_index_col]
        else:
            src_cols = [src_col]

        for idx, row in input_frame[src_cols].iterrows():
            row_num += 1
            observable = row[src_col]

            # Use the user-specified index if possible
            if src_index_col:
                idx = row[src_index_col]

            # validate the observable to avoid sending too much junk to VT
            pp_observable = self._validate_observable(
                observable, ioc_type, idx)

            # if the observable is valid, add it to the submission batch
            if pp_observable.observable:
                obs_batch.append(pp_observable.observable)
                source_row_index[pp_observable.observable] = idx
                batch_index += 1

            # We want to trigger in the following circumstances
            # 1. if the length of our batch is at the max VT batchsize for this type
            #   (If the batch size is 1 this will fire for every row)
            # 2. Or we have reached the end of our row iteration
            # AND
            # 3. The batch is not empty
            if ((len(obs_batch) == vt_param.batch_size or row_num == row_count) and
                    obs_batch):
                obs_submit = vt_param.batch_delimiter.join(obs_batch)

                self._print_status(
                    'Submitting observables: "{}", type "{}" to VT. (Source index {})'
                    .format(obs_submit, ioc_type, idx), 2)
                # Submit the request
                results, status_code = self._vt_submit_request(
                    obs_submit, vt_param)

                if status_code != 200:
                    # Print status messages and add failure cases to results
                    status = 'Failed submission: http error {}'.format(
                        status_code)
                    for failed_obs in obs_batch:
                        self._add_invalid_input_result(failed_obs, ioc_type, status,
                                                       source_row_index[failed_obs])
                        self._print_status(
                            'Error in response submitting observables: "{}", type "{}" '
                            'http status is {}. Response: {} (Source index {})'
                            .format(obs_submit, ioc_type, status_code, results,
                                    source_row_index[failed_obs]), 1)
                else:
                    # parse the results from the response
                    self._parse_vt_results(
                        results, obs_submit, ioc_type, idx, source_row_index, vt_param)

                # reset index of batch
                batch_index = 0
                obs_batch = []

    def _parse_vt_results(self, vt_results: any, observable: str, ioc_type: str,
                          source_idx: any = 0,
                          source_row_index: any = None, vt_param: VTParams = None):
        """
        Parse VirusTotal results based on IoCType.

            :param vt_results: Raw results from VT
            :param observable: The observable or observable batch
            :param ioc_type: The IoC type of the observables
            :param source_idx: The row index of the source frame
            :param source_row_index: (batch only) Mapping between observable item
                and row index of the source
            :param vt_param: (batch only) the VTParams tuple for this submission

        """
        results_to_parse = []
        if isinstance(vt_results, str):
            try:
                vt_results = json.loads(vt_results, strict=False)
            except (JSONDecodeError, TypeError):
                pass

        if isinstance(vt_results, list) and vt_param.batch_size > 1:
            # multiple results
            results_to_parse = vt_results
        elif isinstance(vt_results, dict):
            # single result
            results_to_parse.append(vt_results)
        else:
            self._print_status('Error parsing response to JSON: "{}", type "{}". (Source index {})'
                               .format(observable, ioc_type, source_idx), 1)

        if vt_param and vt_param.batch_delimiter:
            observables = observable.split(vt_param.batch_delimiter)
        else:
            observables = [observable]

        # pylint: disable=locally-disabled, C0200
        for result_idx in range(0, len(results_to_parse)):
            df_dict_vtresults = self._parse_single_result(
                results_to_parse[result_idx], ioc_type)

            # Add remaining fields from source
            df_dict_vtresults['IoCType'] = ioc_type
            df_dict_vtresults['Status'] = 'Success'
            df_dict_vtresults['RawResponse'] = json.dumps(
                results_to_parse[result_idx])
            if len(results_to_parse) == 1 or source_row_index is None or len(source_row_index) == 1:
                df_dict_vtresults['Observable'] = observable
                df_dict_vtresults['SourceIndex'] = source_idx
            else:
                # If we submitted multiple values in a batch
                # we assume (hope) that the ordering of the response is the same
                # as in the request. We try our best to remarry the observable
                # and source index
                if 'resource' in results_to_parse[result_idx]:
                    vt_resource = results_to_parse[result_idx]['resource']
                    df_dict_vtresults['Observable'] = vt_resource
                    if vt_resource in source_row_index:
                        df_dict_vtresults['SourceIndex'] = source_row_index[vt_resource]
                    else:
                        df_dict_vtresults['SourceIndex'] = source_row_index[observables[result_idx]]
                else:
                    df_dict_vtresults['Observable'] = observables[result_idx]
                    df_dict_vtresults['SourceIndex'] = source_row_index[observables[result_idx]]

            if pd_version_23():
                new_results = pd.concat(
                    objs=[self.results, df_dict_vtresults], ignore_index=True, sort=False, axis=0)
            else:
                new_results = pd.concat(
                    objs=[self.results, df_dict_vtresults], ignore_index=True, axis=0)

            self.results = new_results
        # pylint enable=locally-disabled, C0200

    def _parse_single_result(self, results_dict: dict({str: any}), ioc_type: str) -> pd.DataFrame:
        """
        Parse VirusTotal single result based on IoCType.

            :param results_dict: Raw results dictionary from VT
            :param ioc_type: The IoC type of the observables

        """
        # create output frame and parse results to intermediate frame
        df_dict_vtresults = {}

        # Parse returned results to our output dataframe depending
        # on the IoC type
        if ioc_type in ['url', 'md5_hash', 'sha1_hash', 'sha256_hash']:
            df_dict_vtresults['ResponseCode'] = results_dict.get(
                'response_code', None)
            df_dict_vtresults['VerboseMsg'] = results_dict.get(
                'verbose_msg', None)
            df_dict_vtresults['ScanId'] = results_dict.get('scan_id', None)
            df_dict_vtresults['Resource'] = results_dict.get('resource', None)
            df_dict_vtresults['Permalink'] = results_dict.get(
                'permalink', None)
            df_dict_vtresults['Positives'] = results_dict.get(
                'positives', None)
            if ioc_type in ['md5_hash', 'sha1_hash', 'sha256_hash']:
                df_dict_vtresults['MD5'] = results_dict.get('md5', None)
                df_dict_vtresults['SHA1'] = results_dict.get('sha1', None)
                df_dict_vtresults['SHA256'] = results_dict.get('sha256', None)

        if ioc_type == 'ipv4' or ioc_type == 'dns':
            df_dict_vtresults['ResponseCode'] = results_dict.get(
                'response_code', None)
            df_dict_vtresults['VerboseMsg'] = results_dict.get(
                'verbose_msg', None)
            # dns and ipv4 have multi-valued 'resolutions' and 'detected_urls' lists
            # of dictionaries
            # This leads to a few horrendous-looking list comprehensions
            # These are essentially pulling out the columns that contain these lists.
            # then using a list comprehension to pull out the value, where the key 'k'
            # is of the required value
            if ioc_type == 'ipv4' and 'resolutions' in results_dict:
                item_list = [item['hostname']
                             for item in results_dict['resolutions'] if 'hostname' in item]
                df_dict_vtresults['ResolvedDomains'] = ', '.join(item_list)
            elif ioc_type == 'dns' and 'resolutions' in results_dict:
                item_list = [item['ip_address']
                             for item in results_dict['resolutions'] if 'ip_address' in item]
                df_dict_vtresults['ResolvedIPs'] = ', '.join(item_list)
            if 'detected_urls' in results_dict:
                item_list = [item['url']
                             for item in results_dict['detected_urls'] if 'url' in item]
                df_dict_vtresults['DetectedUrls'] = ', '.join(item_list)
                # positives are listed per detected_url so we need to pull those our and sum them.
                positives = sum([item['positives']
                                 for item in results_dict['detected_urls']
                                 if 'positives' in item])
                df_dict_vtresults['Positives'] = positives

        return pd.DataFrame(data=df_dict_vtresults, columns=self._RESULT_COLUMNS, index=[0])

    def _validate_observable(self, observable: str, ioc_type: str, idx: any) -> PreProcessResult:
        """
        Validate observable for format and duplicates of existing results.

        Parameters
        ----------
        observable : str
            The observable to be checked
        ioc_type : str
            The IoCType of the observable
        idx : any
            The index of the source row

        Returns
        -------
        PreProcessResult
            The Pre-processed result

        """
        if observable is None or observable.strip() is None:
            status = 'Failed: Empty or missing observable value'
            self._add_invalid_input_result(
                observable, ioc_type, status, idx)
            self._print_status(status + ' (Source index {})'.format(idx), 1)
            return PreProcessResult(None, status)

        # Check that observable is of the correct format for this type
        # and do any cleaning up required
        pp_observable = self._preprocess_observable(observable, ioc_type)
        if pp_observable.observable is None:
            self._add_invalid_input_result(
                observable, ioc_type, pp_observable.status, idx)
            # pylint: disable=locally-disabled, line-too-long
            self._print_status(
                'Invalid observable format: "{}", type "{}", status: {} - skipping. (Source index {})'
                .format(observable, ioc_type, pp_observable.status, idx), 2)
            # pylint: enable=locally-disabled, line-too-long
            return pp_observable

        # Check that we don't already have a result for this
        dup_result = self._check_duplicate_submission(
            observable, ioc_type, idx)
        if dup_result.is_dup:
            # pylint: disable=locally-disabled, line-too-long
            self._print_status(
                'Duplicate observable value detected: "{}", type "{}" status: {} - skipping. (Source index {})'
                .format(observable, ioc_type, dup_result.status, idx), 2)
            return PreProcessResult(None, dup_result.status)

        return pp_observable

    def _check_duplicate_submission(self, observable, ioc_type, source_index) -> DuplicateStatus:
        """
        Check for a duplicate value in existing results.

        If duplicate found, add a new result row that is a reference to
        the duplicated result row
            :param observable: The IoC observable value
            :param ioc_type: The IoC type
            :param source_idx: The index of the source DataFrame row

        Returns:
            DuplicateStatus{bool, str}

        """
        if self.results is None:
            return DuplicateStatus(False, 'ok')

        # Note duplicate var here can be multiple rows of past results
        duplicate = self.results[self.results['Observable'] == observable].copy()
        # if this is a file hash we should check for previous results in all of the hash
        # columns
        if duplicate.shape[0] == 0 and ioc_type in ['md5_hash', 'sha1_hash', 'sh256_hash']:
            dup_query = 'MD5 == @observable or SHA1 == @observable or SHA256 == @observable'
            duplicate = self.results.query(dup_query).copy()
            # In these cases we want to set the observable to the source value but keep the
            # rest of the results
            if duplicate.shape[0] > 0:
                duplicate['Observable'] = observable

        # if we found a duplicate so add the copies of the duplicated requests
        # to the results
        if duplicate.shape[0] > 0:
            original_indices = [v[0]
                                for v in duplicate[['SourceIndex']].values]
            duplicate['SourceIndex'] = source_index
            duplicate['Status'] = 'Duplicate'
            if pd_version_23():
                new_results = pd.concat(
                    objs=[self.results, duplicate], ignore_index=True, sort=False, axis=0)
            else:
                new_results = pd.concat(
                    objs=[self.results, duplicate], ignore_index=True, axis=0)
            self.results = new_results

            return DuplicateStatus(True, 'Duplicates of {}'.format(original_indices))

        return DuplicateStatus(False, 'ok')

    def _add_invalid_input_result(self, observable, ioc_type, status, source_idx):
        """
        Add a result row to indicate an invalid submission.

            :param observable: The IoC observable value
            :param ioc_type: The IoC type
            :param status: The status - why the item was invalid
            :param source_idx: The index of the source DataFrame row

        """
        new_row = pd.Series(index=self._RESULT_COLUMNS)
        new_row['Observable'] = observable
        new_row['IoCType'] = ioc_type
        new_row['Status'] = status
        new_row['SourceIndex'] = source_idx
        new_results = self.results.append(new_row.to_dict(), ignore_index=True)

        self.results = new_results

    def _vt_submit_request(self, submission_string, vt_param):
        """
        Submit the request to VT.

            :param submission_string: The observable (or observable collection)
            :param vt_param: VT parameters appropriate to this observable type
        """
        params = {'apikey': self._vtkey,
                  vt_param.api_var_name: submission_string}
        submit_url = self._get_vt_api_url(vt_param.api_type)
        headers = {'User-Agent': 'VirusTotal',
                   'Content-Type': 'application/json'}
        if vt_param.headers is not None:
            for hdr, val in vt_param.headers.items():
                headers[hdr] = val

        if vt_param.http_verb == 'post':
            response = requests.post(submit_url, data=params, headers=headers)
        else:
            response = requests.get(submit_url, params=params, headers=headers)
        if response.status_code == 200:
            return response.json(), response.status_code
        else:
            if response:
                try:
                    return response.json(), response.status_code
                except JSONDecodeError:
                    pass
            return None, response.status_code

    def _preprocess_observable(self, observable, ioc_type) -> PreProcessResult:
        """
        Preprocesses and checks validity of observable against declared IoC type.

            :param observable: the value of the IoC
            :param ioc_type: the IoC type
        """
        observable = observable.strip()
        if not self._ioc_extract.validate(observable, ioc_type):
            return PreProcessResult(None,
                                    'Observable does not match expected pattern for ' + ioc_type)
        if ioc_type == 'url':
            return self._preprocess_url(observable)
        if ioc_type == 'ipv4':
            return self._preprocess_ip4(observable)
        if ioc_type == 'dns':
            return self._preprocess_dns(observable)
        if ioc_type in ['md5_hash', 'sha1_hash', 'sha256_hash']:
            return self._preprocess_hash(observable)
        return PreProcessResult(observable, 'ok')

    @classmethod
    def _preprocess_url(cls, url: str) -> PreProcessResult:
        """
        Check that URL can be parsed.

            :param cls: the class
            :param url: the URL to check
        """
        try:
            scheme, _, host, _, _, _, _ = parse_url(url)
            clean_url = url
        except LocationParseError:
            # Try to clean URL and re-check
            clean_url = cls._clean_url(url)
            if clean_url is None:
                return PreProcessResult(None, 'Could not parse as valid URL')
            try:
                scheme, _, host, _, _, _, _ = parse_url(clean_url)
            except LocationParseError:
                return PreProcessResult(None, 'Could not parse as valid URL')

        if scheme is None or host is None:
            return PreProcessResult(None, f'url scheme or host missing from {url}')
        # get rid of some obvious false positives (localhost, local hostnames)
        try:
            addr = ip_address(host)
            if addr.is_private:
                return PreProcessResult(None, 'Host part of URL is a private IP address')
            if addr.is_loopback:
                return PreProcessResult(None, 'Host part of URL is a loopback IP address')
        except ValueError:
            pass

        if '.' not in host:
            return PreProcessResult(None, 'Host is unqualified domain name')

        if scheme.lower() in ['file']:
            return PreProcessResult(None, f'{scheme} URL scheme is not supported')

        return PreProcessResult(clean_url, 'ok')

    @classmethod
    def _clean_url(cls, url: str) -> str:
        """
        Clean URL to remove query params and fragments and any trailing stuff.

            :param cls: the class
            :param url: the URL to check
        """
        # slightly stricter than normal URL regex to exclude '() from host string
        http_strict_regex = r'''
            (?P<protocol>(https?|ftp|telnet|ldap|file)://)
            (?P<userinfo>([a-z0-9-._~!$&*+,;=:]|%[0-9A-F]{2})*@)?
            (?P<host>([a-z0-9-._~!$&\*+,;=]|%[0-9A-F]{2})*)
            (:(?P<port>\d*))?
            (/(?P<path>([^?\#| ]|%[0-9A-F]{2})*))?
            (\?(?P<query>([a-z0-9-._~!$&'()*+,;=:/?@]|%[0-9A-F]{2})*))?
            (\#(?P<fragment>([a-z0-9-._~!$&'()*+,;=:/?@]|%[0-9A-F]{2})*))?\b'''
        if cls._http_strict_rgxc is None:
            cls._http_strict_rgxc = re.compile(
                http_strict_regex, re.I | re.X | re.M)

        # Try to clean URL and re-check
        match_url = cls._http_strict_rgxc.search(url)
        if (match_url.groupdict()['protocol'] is None or
                match_url.groupdict()['host'] is None):
            return None, 'Could not parse as valid URL'

        # build the URL dropping the query string and fragments
        clean_url = match_url.groupdict()['protocol']
        if match_url.groupdict()['userinfo']:
            clean_url += match_url.groupdict()['userinfo']
        clean_url += match_url.groupdict()['host']
        if match_url.groupdict()['port']:
            clean_url += ':' + match_url.groupdict()['port']
        if match_url.groupdict()['path']:
            clean_url += '/' + match_url.groupdict()['path']

        return clean_url

    @classmethod
    def _preprocess_ip4(cls, ipaddress: str):
        """Ensure Ip address is a valid public IPv4 address."""
        try:
            addr = ip_address(ipaddress)
        except ValueError:
            return PreProcessResult(None, 'IP address is invalid format')

        if not isinstance(addr, IPv4Address):
            return PreProcessResult(None, 'Not an IPv4 address')
        if addr.is_global:
            return PreProcessResult(ipaddress, 'ok')
        if addr.is_private:
            return PreProcessResult(None, 'IP is private address')
        if addr.is_loopback:
            return PreProcessResult(None, 'IP is loopback address')
        if addr.is_reserved:
            return PreProcessResult(None, 'IP is reserved address')
        if addr.is_multicast:
            return PreProcessResult(None, 'IP is multicast address')
        return PreProcessResult(None, 'IP address is not global')

    @classmethod
    def _preprocess_dns(cls, domain: str) -> PreProcessResult:
        """Ensure DNS is a valid-looking domain."""
        if '.' not in domain:
            return PreProcessResult(None, 'Domain is unqualified domain name')
        try:
            addr = ip_address(domain)
            del addr
            return PreProcessResult(None, 'Domain is an IP address')
        except ValueError:
            pass
        try:
            _ = socket.gethostbyname(domain)
        except gaierror:
            return PreProcessResult(None, 'Domain not resolvable')

        return PreProcessResult(domain, 'ok')

    @classmethod
    def _preprocess_hash(cls, hash_str: str) -> PreProcessResult:
        """Ensure Hash has minimum entropy (rather than a string of 'x')."""
        str_entropy = cls.entropy(hash_str)
        if str_entropy < 3.0:
            return PreProcessResult(None, 'String has too low an entropy to be a hash')
        return PreProcessResult(hash_str, 'ok')

    @classmethod
    def entropy(cls, input_str: str) -> float:
        """Compute entropy of input string."""
        str_len = float(len(input_str))
        return -sum(map(lambda a: (a / str_len) * math.log2(a / str_len),
                        Counter(input_str).values()))

    @classmethod
    def _get_vt_api_url(cls, api_type: str) -> str:
        """
        Return the VirusTotal API URL for the supplied type.

            :param api_type: The IoC type
        """
        if api_type not in cls._VT_API_TYPES:
            raise LookupError('Unknown api type "{}"'.format(api_type))
        return cls._VT_API.format(type=api_type)

    @classmethod
    def _get_supported_vt_ioc_types(cls) -> list({str}):
        """Return the subset of IoC types supported by VT."""
        return [t for t in cls._SUPPORTED_INPUT_TYPES if cls._VT_TYPE_MAP[t] is not None]

    def _print_status(self, message: str, verbosity_level: str):
        """
        Print a status message depending on the current level of verbosity.

            :param message: the string message to print
            :param verbosity_level: at which level the message should be output
        """
        if verbosity_level <= self._verbosity:
            print(message)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Outlier detection class. TODO **Preliminary**."""

import math

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from sklearn.ensemble import IsolationForest

from .. _version import VERSION

__version__ = VERSION
__author__ = 'Ian Hellen'

# pylint: disable=all
# noqa


def identify_outliers(X: np.array, X_predict: np.Array = None, contamination: float = 0.05):
    """
    Identify outlier items using SkLearn IsolationForest.

    Arguments:
        X {[np.array]} -- Input data

    Keyword Arguments:
        X_predict {[np.array]} -- Model (default: {None})
        contamination {float} -- Percentage contamination (default: {0.05})
    """
    rng = np.random.RandomState(42)

    # fit the model
    rows, cols = X.shape
    max_samples = min(100, cols)
    max_features = math.floor(math.sqrt(rows))
    clf = IsolationForest(max_samples=max_samples, max_features=max_features,
                          random_state=rng, contamination=contamination)
    clf.fit(X)
    y_pred_train = clf.predict(X)

    y_pred_outliers = clf.predict(X_predict)

    X_outliers = X_predict[y_pred_outliers == -1]
    return clf, X_outliers, y_pred_outliers


def plot_outlier_results(clf: IsolationForest, X: np.array,
                         X_predict: np.array, X_outliers: np.array,
                         feature_columns: list({int}), plt_title: str):
    """
    Plot Isolation Forest results.

    Arguments:
        clf {IsolationForest} -- Isolation Forest model
        X {np.array} -- Input data
        X_predict {np.array} -- Prediction
        X_outliers {np.array} -- Set of outliers
        feature_columns {list} -- list of feature columns to display
        plt_title {str} -- Title
    """
    # plot the line, the samples, and the nearest vectors to the plane
    x_max_x = X[:, 0].max() + (X[:, 0].max() / 10)
    x_min_x = - X[:, 0].max() / 10
    x_max_y = X[:, 1].max() + (X[:, 1].max() / 10)
    x_min_y = - X[:, 1].max() / 10
    xx, yy = np.meshgrid(np.linspace(x_min_x, x_max_x, 100),
                         np.linspace(x_min_y, x_max_y, 100))
    Z = clf.decision_function(np.c_[xx.ravel(), yy.ravel()])
    Z = Z.reshape(xx.shape)

    plt.rcParams['figure.figsize'] = (20, 10)

    plt.title(plt_title)
    plt.contourf(xx, yy, Z, cmap=plt.cm.Blues_r)

    b1 = plt.scatter(X[:, 0], X[:, 1], c='white',
                     s=20, edgecolor='k')
    b2 = plt.scatter(X_predict[:, 0], X_predict[:, 1], c='green',
                     s=40, edgecolor='k')
    c = plt.scatter(X_outliers[:, 0], X_outliers[:, 1], c='red', marker='x',
                    s=200, edgecolor='k')
    plt.axis('tight')

    xp_max_x = X_predict[:, 0].max() + (X_predict[:, 0].max() / 10)
    xp_min_x = - X_predict[:, 0].max() / 10
    xp_max_y = X_predict[:, 1].max() + (X_predict[:, 1].max() / 10)
    xp_min_y = - X_predict[:, 1].max() / 10

    plt.xlim((xp_min_x, xp_max_x))
    plt.ylim((xp_min_y, xp_max_y))
    plt.xlabel(feature_columns[0])
    plt.ylabel(feature_columns[1])

    plt.legend([b1, b2, c],
               ["training observations",
                "new regular observations", "new abnormal observations"],
               loc="upper right")
    plt.show()


def remove_common_items(data: pd.DataFrame, columns: list):
    """
    Remove rows from input DataFrame.

    Arguments:
        data {pd.DataFrame} -- [description]
        columns {list} -- [description]
    """
    filtered_df = data
    for col in columns:
        filtered_df = filtered_df.filter(lambda x: (x[col].std() == 0 and
                                                    x[col].count() > 10))

    return filtered_df



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Folium map class."""
from collections.abc import Iterable
from numbers import Number
import warnings

import folium

# pylint: enable=locally-disabled, W0611
from . utility import export
from .. _version import VERSION

__version__ = VERSION
__author__ = 'Ian Hellen'


@export
class FoliumMap():
    """Wrapper class for Folium/Leaflet mapping."""

    def __init__(self, title: str = 'layer1', zoom_start: int = 7, tiles=None,
                 width: str = '100%', height: str = '100%'):
        """
        Create an instance of the folium map.

        Keyword Arguments:
            title {str} -- Name of the layer (default: {'layer1'})
            zoom_start {int} -- The zoom level of the map (default: {7})
            tiles {[type]} -- Custom set of tiles or tile URL (default: {None})
            width {str} -- Map display width (default: {'100%'})
            height {str} -- Map display height (default: {'100%'})

        """
        self.folium_map = folium.Map(
            zoom_start=zoom_start, tiles=tiles, width=width, height=height)
        self.folium_map.add_tile_layer(name=title)

    def add_ip_cluster(self, ip_entities: Iterable, **kwargs):
        """
        Add a collection of IP Entities to the map.

        Arguments:
            ip_entities {Iterable} -- a list of IpAddress Entities

        Keyword Arguments:
            kwargs: icon properties to use for displaying this cluster

        """
        for ip_entity in ip_entities:
            if not (isinstance(ip_entity.Location.Latitude, Number) and
                    isinstance(ip_entity.Location.Longitude, Number)):
                warnings.warn("Invalid location information for IP: " + ip_entity.Address,
                              RuntimeWarning)
                continue
            loc_props = ', '.join([f'{key}={val}' for key, val in
                                   ip_entity.Location.properties.items() if val])
            popup_text = "{loc_props}<br>{IP}".format(IP=ip_entity.Address,
                                                      loc_props=loc_props)
            tooltip_text = '{City}, {CountryName}'.format(
                **ip_entity.Location.properties)

            if ip_entity.AdditionalData:
                addl_props = ', '.join([f'{key}={val}' for key, val in
                                        ip_entity.AdditionalData.items() if val])
                popup_text = f'{popup_text}<br>{addl_props}'
                tooltip_text = f'{tooltip_text}, {addl_props}'
            marker = folium.Marker(
                location=[ip_entity.Location.Latitude,
                          ip_entity.Location.Longitude],
                popup=popup_text,
                tooltip=tooltip_text,
                icon=folium.Icon(**kwargs)
            )
            marker.add_to(self.folium_map)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for common display functions."""

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
from bokeh.io import output_notebook, show
from bokeh.models import (ColumnDataSource, DatetimeTickFormatter, HoverTool,
                          Label)
from bokeh.plotting import figure, reset_output
from IPython.core.display import HTML, display
from IPython.display import Javascript

from .security_alert import SecurityAlert
from .utility import export
from .. _version import VERSION

__version__ = VERSION
__author__ = 'Ian Hellen'


@export
def display_alert(alert=None, show_entities=False):
    """
    Display the alert properties as HTML.

        :param alert: The alert to display
            pd.Series or SecurityAlert
    """
    if alert is None:
        return

    if isinstance(alert, SecurityAlert):
        display(HTML(alert.to_html(show_entities=False)))
        if show_entities:
            for entity in alert.entities:
                print(entity)
        return

    # Display subset of raw properties
    if isinstance(alert, pd.Series):
        entity = (alert['CompromisedEntity']
                  if 'CompromisedEntity' in alert
                  else '')
        title = '''
            <h3>Alert: '{name}'</h3><br>time=<b>{start}</b>, entity=<b>{entity}</b>, id=<b>{id}</b>
            '''.format(start=alert['StartTimeUtc'],
                       name=alert['AlertDisplayName'],
                       entity=entity,
                       id=alert['SystemAlertId'])
        display(HTML(title))
        display(pd.DataFrame(alert))
    else:
        raise ValueError(
            'Unrecognized alert object type ' + str(type(alert)))


def _print_process(process_row: pd.Series):
    if process_row.NodeRole == 'parent':
        if process_row.Level > 1:
            level = 0
        else:
            level = 1
    elif process_row.NodeRole == 'source':
        level = 2
    elif process_row.NodeRole == 'child':
        level = 3 + process_row.Level
    else:
        level = 2

    spaces = 20 * level * 2
    if process_row.NodeRole == 'source':
        line1 = '''
        <span style="color:red;font-weight:bold">[alert:lev{}] {} {}
        [PID: {}, SubjSess:{}, TargSess:{}]</span>
        '''.format(process_row.Level,
                   process_row.TimeCreatedUtc,
                   process_row.NewProcessName,
                   process_row.NewProcessId,
                   process_row.SubjectLogonId,
                   process_row.TargetLogonId)
    else:
        line1 = '[{}:lev{}] {} <b>{}</b> [PID: {}, SubjSess:{}, TargSess:{}]'.format(
            process_row.NodeRole,
            process_row.Level,
            process_row.TimeCreatedUtc,
            process_row.NewProcessName,
            process_row.NewProcessId,
            process_row.SubjectLogonId,
            process_row.TargetLogonId)

    line2 = '(Cmdline: \'{}\') [Account: \'{}\']'.format(
        process_row.CommandLine, process_row.SubjectUserName)

    display(HTML('<div style="margin-left:{indent}px">{txt}<br>{txt2}</div>'.format(indent=spaces,
                                                                                    txt=line1,
                                                                                    txt2=line2)))


@export
def display_process_tree(process_tree: pd.DataFrame):
    """
    Display process tree data frame.

        :param process_tree
    """
    tree = process_tree[['TimeCreatedUtc', 'NodeRole', 'Level', 'NewProcessName',
                         'CommandLine', 'SubjectUserName', 'NewProcessId', 'ProcessId',
                         'SubjectLogonId', 'TargetLogonId']]
    tree = tree.sort_values(by=['TimeCreatedUtc'], ascending=False)

    display(HTML("<h3>Alert process tree:</h3>"))
    tree.sort_values(by=['TimeCreatedUtc']).apply(_print_process, 1)


@export
def exec_remaining_cells():
    """Execute all cells below currently selected cell."""
    Javascript("Jupyter.notebook.execute_cells_below()")


@export
def draw_alert_entity_graph(nx_graph: nx.Graph, font_size: int = 12,
                            height: int = 15, width: int = 15,
                            margin: float = 0.3, scale: int = 1):
    """
    "Draw networkX graph with matplotlib.

    Arguments:
        nx_graph {networkx.graph} -- [description]

    Keyword Arguments:
        font_size {int} -- base font size (default: {12})
        height {int} -- Image height (default: {15})
        width {int} -- Image width (default: {20})
        margin {float} -- Image margin (default: {0.3})
        scale {int} -- Position scale (default: {1})
    """
    alert_node = [n for (n, node_type) in
                  nx.get_node_attributes(nx_graph, 'node_type').items()
                  if node_type == 'alert']
    entity_nodes = [n for (n, node_type) in
                    nx.get_node_attributes(nx_graph, 'node_type').items()
                    if node_type == 'entity']

    # now draw them in subsets  using the `nodelist` arg
    plt.rcParams['figure.figsize'] = (width, height)

    plt.margins(x=margin, y=margin)

    pos = nx.kamada_kawai_layout(nx_graph, scale=scale, weight='weight')
    nx.draw_networkx_nodes(nx_graph, pos, nodelist=alert_node,
                           node_color='red', alpha=0.5, node_shape='o')
    nx.draw_networkx_nodes(nx_graph, pos, nodelist=entity_nodes,
                           node_color='green', alpha=0.5, node_shape='s',
                           s=200)
    nlabels = nx.get_node_attributes(nx_graph, 'description')
    nx.relabel_nodes(nx_graph, nlabels)
    nx.draw_networkx_labels(nx_graph, pos, nlabels, font_size=font_size)
    nx.draw_networkx_edges(nx_graph, pos)
    elabels = nx.get_edge_attributes(nx_graph, 'description')
    nx.draw_networkx_edge_labels(nx_graph, pos, edge_labels=elabels,
                                 font_size=font_size * 2 / 3, alpha=0.6)


@export
def display_timeline(data, alert=None, overlay_data=None, title: str = None,
                     time_column: str = 'TimeGenerated',
                     source_columns: list = None,
                     overlay_colums: list = None,
                     height: int = 300):
    """
    Display a timeline of events.

    Arguments:
        data {pd.DataFrame} -- Input DataFrame

    Keyword Arguments:
        alert {SecurityAlert} -- Input alert (optional) (default: {None})
        overlay_data {pd.DataFrame} -- Second event stream (DataFrame)
            to display as overlay (default: {None})
        title {str} -- [description] (default: {None})
        time_column {str} -- The name of the time
            property used in the Dataframe(s) (default: {'TimeGenerated'})
        source_columns {list} -- List of source columns to use in
            tooltips (default: {None})
        overlay_colums {list} -- List of source columns to use in
            overlay data tooltips (default: {None})
        heigh {int} -- the height of the plot figure (under 300 limits access
            to Bokeh tools)
    """
    reset_output()
    output_notebook()

# pylint: disable=C0103
    WRAP = 50
    WRAP_CMDL = 'WrapCmdl'
# pylint: enable=C0103
    y_max = 1

    if not source_columns:
        source_columns = ['NewProcessName', 'EventID', 'CommandLine']
    if time_column not in source_columns:
        source_columns.append(time_column)

    if 'CommandLine' in source_columns:
        graph_df = data[source_columns].copy()
        graph_df[WRAP_CMDL] = graph_df.apply(lambda x:
                                             _wrap_text(x.CommandLine, WRAP),
                                             axis=1)
    else:
        graph_df = data[source_columns].copy()

    # if we have an overlay - add this data and shift the y co-ordinates to
    # show on two separate lines
    if overlay_data is not None:
        overlay_colums = (overlay_colums if overlay_colums is not None
                          else source_columns)
        if time_column not in overlay_colums:
            overlay_colums.append(time_column)
        if 'CommandLine' in overlay_colums:
            overlay_df = overlay_data[overlay_colums].copy()
            overlay_df[WRAP_CMDL] = overlay_df.apply(lambda x:
                                                     _wrap_text(
                                                         x.CommandLine, WRAP),
                                                     axis=1)
        else:
            overlay_df = overlay_data[overlay_colums].copy()
        graph_df['y_index'] = 2
        overlay_df['y_index'] = 1
        y_max = 2
    else:
        graph_df['y_index'] = 1

    source = ColumnDataSource(graph_df)

    # build the tool tips from columns (excluding these)
    excl_cols = [time_column, 'CommandLine']
    tool_tip_items = [(f'{col}', f'@{col}')
                      for col in source_columns if col not in excl_cols]
    if WRAP_CMDL in graph_df:
        tool_tip_items.append(('CommandLine', f'@{WRAP_CMDL}'))
    hover = HoverTool(
        tooltips=tool_tip_items,
        formatters={'Tooltip': 'printf'}
        # display a tooltip whenever the cursor is vertically in line with a glyph
        # ,mode='vline'
    )

    if not title:
        title = 'Event Timeline'
    else:
        title = 'Timeline {}'.format(title)

    # tools = 'pan, box_zoom, wheel_zoom, reset, undo, redo, save, hover'
    plot = figure(min_border_left=50, plot_height=height, plot_width=900,
                  x_axis_label='Event Time', x_axis_type='datetime', x_minor_ticks=10,
                  tools=[hover, 'pan', 'xwheel_zoom', 'box_zoom', 'reset'],
                  title=title)
    plot.yaxis.visible = False

    # Tick formatting for different zoom levels
    # '%H:%M:%S.%3Nms
    tick_format = DatetimeTickFormatter()
    tick_format.days = ['%m-%d %H:%M']
    tick_format.hours = ['%H:%M:%S']
    tick_format.minutes = ['%H:%M:%S']
    tick_format.seconds = ['%H:%M:%S']
    tick_format.milliseconds = ['%H:%M:%S.%3N']

    plot.xaxis[0].formatter = tick_format
    plot.circle(x=time_column, y='y_index', color='navy',
                alpha=0.5, size=10, source=source)

    if overlay_data is not None:
        overlay_source = ColumnDataSource(overlay_df)
        plot.circle(x=time_column, y='y_index', color='green',
                    alpha=0.5, size=10, source=overlay_source)

    # Adding data labels stops everything working!
    # labels = LabelSet(x=time_column, y='y_index', y_offset=5,
    #                   text='NewProcessName', source=source,
    #                   angle='90deg', text_font_size='8pt')
    # p.add_layout(labels)

    # if we have an alert, plot the time as a line
    if alert is not None:
        x_alert_label = pd.Timestamp(alert['StartTimeUtc'])
        plot.line(x=[x_alert_label, x_alert_label], y=[0, y_max + 1])
        alert_label = Label(x=x_alert_label, y=0, y_offset=10, x_units='data', y_units='data',
                            text='< Alert time', render_mode='css',
                            border_line_color='red', border_line_alpha=1.0,
                            background_fill_color='white', background_fill_alpha=1.0)

        plot.add_layout(alert_label)

        print('Alert start time = ', alert['StartTimeUtc'])

    show(plot)


def _wrap_text(source_string, wrap_len):
    if len(source_string) <= wrap_len:
        return source_string
    out_string = ''
    input_parts = source_string.split()
    out_line = ''
    for part in input_parts:
        if len(part) > wrap_len:
            if out_line:
                out_string += out_line + '\n'
                out_line = ''
            out_line = part[0:wrap_len] + '...'
        else:
            if out_line:
                out_line += ' ' + part
            else:
                out_line = part
            if len(out_line) > wrap_len:
                out_string += out_line + '\n'
                out_line = ''

    return out_string


# Constants for Windows logon
_WIN_LOGON_TYPE_MAP = {0: 'Unknown',
                       2: 'Interactive', 3: 'Network', 4: 'Batch', 5: 'Service',
                       7: 'Unlock', 8: 'NetworkCleartext', 9: 'NewCredentials',
                       10: 'RemoteInteractive', 11: 'CachedInteractive'}
_WINDOWS_SID = {'S-1-0-0': 'Null SID', 'S-1-5-18': 'LOCAL_SYSTEM',
                'S-1-5-19': 'LOCAL_SERVICE', 'S-1-5-20': 'NETWORK_SERVICE'}
_ADMINISTRATOR_SID = '500'
_GUEST_SID = '501'
_DOM_OR_MACHINE_SID = 'S-1-5-21'


@export
def display_logon_data(logon_event: pd.DataFrame, alert: SecurityAlert = None,
                       os_family: str = None):
    """
    Display logon data for one or more events.

    Arguments:
        :logon_event: Dataframe containing one or more logon events
        :security alert: obtain os_family from the security alert
        :os_family: explicitly specify os_family (Linux or Windows)

    """
    if not os_family:
        os_family = alert.os_family if alert else 'Windows'

    for _, logon_row in logon_event.iterrows():
        print('### Account Logon')
        print('Account: ', logon_row['TargetUserName'])
        print('Account Domain: ', logon_row['TargetDomainName'])
        print('Logon Time: ', logon_row['TimeGenerated'])

        if os_family == 'Windows':
            logon_type = logon_row['LogonType']
            logon_desc_idx = logon_type
            if logon_type not in _WIN_LOGON_TYPE_MAP:
                logon_desc_idx = 0
            print(f'Logon type: {logon_type} ({_WIN_LOGON_TYPE_MAP[logon_desc_idx]})')

        account_id = logon_row.TargetUserSid
        print('User Id/SID: ', account_id)
        if os_family == 'Windows':
            _print_sid_info(account_id)
        else:
            print('Audit user: ', logon_row['audit_user'])

        session_id = logon_row['TargetLogonId']
        print(f'Session id \'{session_id}\'', end='  ')
        if session_id == '0x3e7' or session_id == '-1':
            print('System logon session')

        print()
        domain = logon_row['SubjectDomainName']
        if not domain:
            subj_account = logon_row.SubjectUserName
        else:
            subj_account = f'{domain}/{logon_row.SubjectUserName}'
        print('Subject (source) account: ', subj_account)

        print('Logon process: ', logon_row['LogonProcessName'])
        print('Authentication: ', logon_row['AuthenticationPackageName'])
        print('Source IpAddress: ', logon_row['IpAddress'])
        print('Source Host: ', logon_row['WorkstationName'])
        print('Logon status: ', logon_row['Status'])
        print()


def _print_sid_info(sid):
    if sid in _WINDOWS_SID:
        print('    SID {} is {}'.format(sid, _WINDOWS_SID[sid]))
    elif sid.endswith(_ADMINISTRATOR_SID):
        print('    SID {} is administrator'.format(sid))
    elif sid.endswith(_GUEST_SID):
        print('    SID {} is guest'.format(sid))
    if sid.startswith(_DOM_OR_MACHINE_SID):
        print('    SID {} is local machine or domain account'.format(sid))


@export
def plot_cluster(db_cluster, data, X, plot_label=None, plot_features=[0, 1], verbose=False,
                 cut_off=3, xlabel=None, ylabel=None):
    """
    Plot clustered data as scatter chart

    Arguments:
        db_cluster {[type]} -- DBScan Cluster (from SkLearn DBSCAN)
        data {[type]} -- Dataframe containing original data
        X {[type]} -- The DBSCAN predict numpy array

    Keyword Arguments:
        plot_label {str} -- If set the column to use to label data points
            (default: {None})
        plot_features {list} -- [description] Which two features in X to plot
        verbose {bool} -- Verbose execution with some extra info (default: {False})
        cut_off {int} -- The cluster size below which items are considered
            outliers (default: {3})
        xlabel {[type]} -- x-axis label (default: {None})
        ylabel {[type]} -- y-axis label (default: {None})
    """
    if plot_features[0] >= X.shape[1]:
        raise ValueError("plot_features[0] index must be a value from 0 to {}."
                         .format(X.shape[1] - 1))
    if plot_features[1] >= X.shape[1]:
        raise ValueError("plot_features[1] index must be a value from 0 to {}."
                         .format(X.shape[1] - 1))
    if plot_features[0] == plot_features[1]:
        raise ValueError("plot_features indexes must be 2 different values in range 0 to {}."
                         .format(X.shape[1] - 1))

    labels = db_cluster.labels_
    core_samples_mask = np.zeros_like(labels, dtype=bool)
    core_samples_mask[db_cluster.core_sample_indices_] = True
    unique_labels = set(labels)
    colors = [plt.cm.Spectral(each)
              for each in np.linspace(0, 1, len(unique_labels))]
    # Number of clusters in labels, ignoring noise if present.
    n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
    n_noise_ = list(labels).count(-1)
    _, counts = np.unique(labels, return_counts=True)

    if verbose:
        print('Estimated number of clusters: %d' % n_clusters_)
        print('Estimated number of noise points: %d' % n_noise_)
        # print("Silhouette Coefficient: %0.3f"
        #       % metrics.silhouette_score(X, labels))

    if not isinstance(data, pd.DataFrame):
        plot_label = None
    elif plot_label is not None and plot_label not in data:
        plot_label = None

    p_label = None
    for cluster_id, color in zip(unique_labels, colors):
        if cluster_id == -1:
            # Black used for noise.
            color = [0, 0, 0, 1]
        class_member_mask = (labels == cluster_id)

        cluster_size = counts[cluster_id]
        marker_size = cluster_size
        marker = 'o'
        font_size = 'small'
        alpha = 0.4

        if cluster_size < cut_off:
            marker = '+'
            marker_size = 10
            font_size = 'large'
            alpha = 1.0
        first_row = data[class_member_mask].iloc[0]
        xy = X[class_member_mask & core_samples_mask]
        plt.plot(xy[:, plot_features[0]], xy[:, plot_features[1]], marker,
                 markerfacecolor=tuple(color),
                 markersize=marker_size)

        if plot_label:
            if len(first_row) > 0 and plot_label in first_row:
                p_label = first_row[plot_label]
                try:
                    plt.annotate(s=p_label, xy=(xy[0, plot_features[0]], xy[0, plot_features[1]]),
                                 fontsize=font_size, alpha=alpha)
                except IndexError:
                    pass

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title('Estimated number of clusters: %d' % n_clusters_)
    plt.show()
    return plt




# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""geoip module using IPStack and Maxmind GeoLite2."""
# import gzip
from json import JSONDecodeError
import math
import os
from abc import ABCMeta, abstractmethod
from collections.abc import Iterable
from datetime import datetime, timedelta
from typing import Tuple, List, Mapping
from IPython import get_ipython
from IPython.display import display, HTML

import pandas as pd
import requests

from geolite2 import geolite2

from .. nbtools.entityschema import GeoLocation, IpAddress
from .. nbtools.utility import export
from .. _version import VERSION

__version__ = VERSION
__author__ = 'Ian Hellen'


class GeoIpLookup(metaclass=ABCMeta):
    """
    Abstract base class for GeoIP Lookup classes.

    Parameters
    ----------
    metaclass : ABCMeta, optional
        Defines class as abstract base class (the default is ABCMeta)

    See Also:
        IPStackLookup : IPStack GeoIP Implementation
        GeoLiteLookup : MaxMind GeoIP Implementation

    """

    @abstractmethod
    def lookup_ip(self, ip_address: str = None,
                  ip_addr_list: Iterable = None,
                  ip_entity: IpAddress = None) -> Tuple[List[Mapping[str, str]], List[IpAddress]]:
        """
        Lookup IP location abstract method.

        Parameters
        ----------
        ip_address : str, optional
            a single address to look up (the default is None)
        ip_addr_list : Iterable, optional
            a collection of addresses to lookup (the default is None)
        ip_entity : IpAddress, optional
            an IpAddress entity (the default is None) - any existing
            data in the Location property will be overwritten

        Returns
        -------
        Tuple[List[Mapping[str, str]], List[IpAddress]]
            raw geolocation results and same results as IpAddress entities with
            populated Location property.

        """

    def df_lookup_ip(self, data: pd.DataFrame, column: str) -> pd.DataFrame:
        """
        Lookup Geolocation data from a pandas Dataframe.

        Parameters
        ----------
        data : pd.DataFrame
            pandas dataframe containing IpAddress column
        column : str
            the name of the dataframe column to use as a source

        Returns
        -------
        pd.DataFrame
            Copy of original dataframe with IP Location information columns
            appended (where a location lookup was successful)

        """
        ip_list = data[column].values
        _, entities = self.lookup_ip(ip_addr_list=ip_list)

        ip_dicts = [ent.Location.properties().update(IpAddress=ent.Address)
                    for ent in entities]
        df_out = pd.DataFrame(data=ip_dicts)
        return data.merge(df_out, how='left', left_on=column, right_on='IpAddress')


@export
class IPStackLookup(GeoIpLookup):
    """
    IPStack GeoIP Implementation.

    Parameters
    ----------
    GeoIpLookup : GeoIPLookup
        Abstract base class for IPLookup implementations

    See Also:
        GeoIpLookup : Abstract base class
        GeoLiteLookup : MaxMind GeoIP Implementation

    """

    _IPSTACK_API = 'http://api.ipstack.com/{iplist}?access_key={access_key}&output=json'

    def __init__(self, api_key: str, bulk_lookup: bool = False):
        """
        Create a new instance of IPStackLookup.

        Parameters
        ----------
        api_key : str
            API Key from IPStack - see https://ipstack.com
        bulk_lookup : bool, optional
            For Professional and above tiers allowing you to
            submit multiple IPs in a single request.
            (the default is False, which submits a single request
            per address)

        """
        self._api_key = api_key
        self.bulk_lookup = bulk_lookup

    def lookup_ip(self, ip_address: str = None,
                  ip_addr_list: Iterable = None,
                  ip_entity: IpAddress = None) -> Tuple[List[Mapping[str, str]], List[IpAddress]]:
        """
        Lookup IP location from IPStack web service.

        Parameters
        ----------
        ip_address : str, optional
            a single address to look up (the default is None)
        ip_addr_list : Iterable, optional
            a collection of addresses to lookup (the default is None)
        ip_entity : IpAddress, optional
            an IpAddress entity (the default is None) - any existing
            data in the Location property will be overwritten

        Returns
        -------
        Tuple[List[Mapping[str, str]], List[IpAddress]]
            raw geolocation results and same results as IpAddress entities with
            populated Location property.

        Raises
        ------
        ConnectionError
            Invalid status returned from http request
        PermissionError
            Service refused request (e.g. requesting batch of addresses
            on free tier API key)

        """
        if ip_address and isinstance(ip_address, str):
            ip_list = [ip_address.strip()]
        elif ip_addr_list:
            ip_list = (ip.strip() for ip in ip_addr_list)
        elif ip_entity:
            ip_list = [ip_entity.Address]
        else:
            raise ValueError('No valid ip addresses were passed as arguments.')

        results = self._submit_request(ip_list)
        output_entities = []
        for ip_loc, status in results:
            if status == 200:
                output_entities.append(
                    self._create_ip_entity(ip_loc, ip_entity))

        return results, output_entities

    def _create_ip_entity(self, ip_loc: dict, ip_entity):
        if not ip_entity:
            ip_entity = IpAddress()
            ip_entity.Address = ip_loc['ip']
        geo_entity = GeoLocation()
        geo_entity.CountryCode = ip_loc['country_code']

        geo_entity.CountryName = ip_loc['country_name']
        geo_entity.State = ip_loc['region_name']
        geo_entity.City = ip_loc['city']
        geo_entity.Longitude = ip_loc['longitude']
        geo_entity.Latitude = ip_loc['latitude']
        if 'connection' in ip_loc:
            geo_entity.Asn = ip_loc['connection']['asn']
        ip_entity.Location = geo_entity
        return ip_entity

    def _submit_request(self, ip_list: List[str]) -> List[Tuple[str, int]]:
        """
        Submit the request to IPStack.

        Parameters
        ----------
        ip_list : List[str]
            String list of IPs to look up

        Returns
        -------
        List[Tuple[str, int]]
            List of response, status code pairs

        """
        ip_loc_results = []
        if not self.bulk_lookup:
            for ip_addr in ip_list:
                submit_url = self._IPSTACK_API.format(
                    iplist=ip_addr, access_key=self._api_key)
                response = requests.get(submit_url)
                if response.status_code == 200:
                    ip_loc_results.append(
                        (response.json(), response.status_code))
                else:
                    if response:
                        try:
                            ip_loc_results.append(
                                (response.json(), response.status_code))
                            continue
                        except JSONDecodeError:
                            ip_loc_results.append(None, response.status_code)
                    else:
                        print('Unknown response from IPStack request.')
                        ip_loc_results.append(None, -1)
            return ip_loc_results
        else:
            submit_url = self._IPSTACK_API.format(iplist=','.join(ip_list),
                                                  access_key=self._api_key)
            response = requests.get(submit_url)

            if response.status_code == 200:
                results = response.json()
                # {"success":false,"error":{"code":303,"type":"batch_not_supported_on_plan",
                # "info":"Bulk requests are not supported on your plan.
                # Please upgrade your subscription."}}

                if 'success' in results and not results["success"]:
                    raise PermissionError('Service unable to complete request. Error: {}'
                                          .format(results['error']))
                return [(item, response.status_code) for item in results]
            else:
                if response:
                    try:
                        return [(response.json(), response.status_code)]
                    except JSONDecodeError:
                        pass
                return [(None, response.status_code)]


@export
class GeoLiteLookup(GeoIpLookup):
    """
    GeoIP Lookup using MaxMindDB database.

    Parameters
    ----------
    GeoIpLookup : GeoIPLookup
        Abstract base class for IPLookup implementations

    See Also:
        IPStackLookup : IPStack GeoIP Implementation
        GeoLiteLookup : MaxMind GeoIP Implementation

    """

    _MAXMIND_DOWNLOAD = 'https://dev.maxmind.com/geoip/geoip2/geolite2/#Downloads'
    _DB_FILE = 'GeoLite2-City.mmdb'

    def __init__(self):
        """Return new instance of GeoLiteLookup class."""
        self._reader = geolite2.reader()
        last_mod_time = datetime.fromtimestamp(
            os.path.getmtime(geolite2.filename))
        db_age = datetime.utcnow() - last_mod_time
        if db_age > timedelta(40):
            print(
                f'{self._DB_FILE} is over one month old. Update the maxminddb package')
            print(
                f'to refresh or download a new version from {self._MAXMIND_DOWNLOAD}')

    def lookup_ip(self, ip_address: str = None,
                  ip_addr_list: Iterable = None,
                  ip_entity: IpAddress = None) -> Tuple[List[Mapping[str, str]], List[IpAddress]]:
        """
        Lookup IP location from IPStack web service.

        Parameters
        ----------
        ip_address : str, optional
            a single address to look up (the default is None)
        ip_addr_list : Iterable, optional
            a collection of addresses to lookup (the default is None)
        ip_entity : IpAddress, optional
            an IpAddress entity (the default is None) - any existing
            data in the Location property will be overwritten

        Returns
        -------
        Tuple[List[Mapping[str, str]], List[IpAddress]]
            raw geolocation results and same results as IpAddress entities with
            populated Location property.

        """
        if ip_address and isinstance(ip_address, str):
            ip_list = [ip_address.strip()]
        elif ip_addr_list:
            ip_list = (ip.strip() for ip in ip_addr_list)
        elif ip_entity:
            ip_list = [ip_entity.Address]
        else:
            raise ValueError('No valid ip addresses were passed as arguments.')

        output_raw = []
        output_entities = []
        for ip_input in ip_list:
            geo_match = self._reader.get(ip_input)
            if geo_match:
                output_raw.append(geo_match)
                output_entities.append(self._create_ip_entity(
                    ip_input, geo_match, ip_entity))

        return output_raw, output_entities

    def _create_ip_entity(self, ip_address, geo_match: dict, ip_entity):
        if not ip_entity:
            ip_entity = IpAddress()
            ip_entity.Address = ip_address
        geo_entity = GeoLocation()
        geo_entity.CountryCode = geo_match.get(
            'country', {}).get('iso_code', None)
        geo_entity.CountryName = geo_match.get(
            'country', {}).get('names', {}).get('en', None)
        subdivs = geo_match.get('subdivisions', [])
        if subdivs:
            geo_entity.State = subdivs[0].get('names', {}).get('en', None)
        geo_entity.City = geo_match.get(
            'city', {}).get('names', {}).get('en', None)
        geo_entity.Longitude = geo_match.get(
            'location', {}).get('longitude', None)
        geo_entity.Latitude = geo_match.get(
            'location', {}).get('latitude', None)
        ip_entity.Location = geo_entity
        return ip_entity


_MM_LICENSE_HTML = '''
This product includes GeoLite2 data created by MaxMind, available from
<a href="https://www.maxmind.com">https://www.maxmind.com</a>.
'''
_MM_LICENSE_TXT = '''
This product includes GeoLite2 data created by MaxMind, available from
https://www.maxmind.com.
'''
_IPSTACK_LICENSE_HTML = '''
This library uses services provided by ipstack.
<a href="https://ipstack.com">https://ipstack.com</a>'''

_IPSTACK_LICENSE_TXT = 'This library uses services provided by ipstack (https://ipstack.com)'

if not get_ipython():
    print(_MM_LICENSE_TXT)
    print(_IPSTACK_LICENSE_TXT)
else:
    display(HTML(_MM_LICENSE_HTML))
    display(HTML(_IPSTACK_LICENSE_HTML))


def entity_distance(ip_src: IpAddress, ip_dest: IpAddress) -> float:
    """
    Return distance between two IP Entities.

    Parameters
    ----------
    ip_src : IpAddress
        Source/Origin IpAddress Entity
    ip_dest : IpAddress
        Destination IpAddress Entity

    Returns
    -------
    float
        Distance in kilometers.

    Raises
    ------
        AttributeError
            If either entity has no location information

    """
    if not ip_src.Location or not ip_dest.Location:
        raise AttributeError(
            'Source and destination entities must have defined Location properties.')

    return geo_distance(origin=(ip_src.Location.Latitude, ip_src.Location.Longitude),
                        destination=(ip_dest.Location.Latitude, ip_dest.Location.Longitude))


_EARTH_RADIUS_KM = 6371  # km


def geo_distance(origin: Tuple[float, float],
                 destination: Tuple[float, float]) -> float:
    """
    Calculate the Haversine distance.

    Parameters
    ----------
    origin : Tuple[float, float]
        Latitude, Longitude of origin of distance measurement.
    destination : Tuple[float, float]
        Latitude, Longitude of origin of distance measurement.

    Returns
    -------
    float
        Distance in kilometers.

    Examples
    --------
    >>> origin = (48.1372, 11.5756)  # Munich
    >>> destination = (52.5186, 13.4083)  # Berlin
    >>> round(geo_distance(origin, destination), 1)
    504.2

    Notes
    -----
        Author: Martin Thoma - stackoverflow

    """
    orig_lat, orig_lon = origin
    dest_lat, dest_lon = destination

    ang_dist_lat = math.radians(dest_lat - orig_lat)
    ang_dist_lon = math.radians(dest_lon - orig_lon)
    hav_a = (math.sin(ang_dist_lat / 2) * math.sin(ang_dist_lat / 2) +
             math.cos(math.radians(orig_lat)) * math.cos(math.radians(dest_lat)) *
             math.sin(ang_dist_lon / 2) * math.sin(ang_dist_lon / 2))
    hav_c = 2 * math.atan2(math.sqrt(hav_a), math.sqrt(1 - hav_a))
    return _EARTH_RADIUS_KM * hav_c



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""default queries module ."""
from . query_defns import KqlQuery, DataFamily, DataEnvironment

# Module level variable that holds dictionary of queries
# indexed by name
# pylint: disable=C0103
query_definitions = dict()
# pylint: enable=C0103


KNOWN_PARAM_NAMES = ['table', 'query_project', 'start', 'end',
                     'system_alert_id', 'subscription_filter',
                     'host_filter_eq', 'host_filter_neq', 'host_name',
                     'account_name', 'process_name', 'process_id',
                     'logon_session_id', 'path_separator', 'commandline',
                     'source_ip_list', 'add_query_items']


def _add_query(kql_query):
    query_definitions[kql_query.name] = kql_query

# ------------------------------------------------------------------------
# Do Not edit above this line
# ------------------------------------------------------------------------


# Predefined queries
_add_query(KqlQuery(name='list_alerts_counts',
                    query='''
{table}
{query_project}
| where TimeGenerated >= datetime({start})
| where TimeGenerated <= datetime({end})
| summarize alertCount=count(), firstAlert=min(TimeGenerated),
    lastAlert=max(TimeGenerated) by AlertName
| order by alertCount desc
{add_query_items}
''',
                    description='Retrieves summary of current alerts',
                    data_source='security_alert',
                    data_families=[DataFamily.SecurityAlert],
                    data_environments=[DataEnvironment.LogAnalytics],
                    optional_params=['add_query_items']))

_add_query(KqlQuery(name='list_alerts',
                    query='''
{table}
{query_project}
| where TimeGenerated >= datetime({start})
| where TimeGenerated <= datetime({end})
| extend extendedProps = parse_json(ExtendedProperties)
| extend CompromisedEntity = tostring(extendedProps['Compromised Host'])
| project-away extendedProps
{add_query_items}
''',
                    description='Retrieves list of current alerts',
                    data_source='security_alert',
                    data_families=[DataFamily.SecurityAlert],
                    data_environments=[DataEnvironment.LogAnalytics],
                    optional_params=['add_query_items']))

_add_query(KqlQuery(name='get_alert',
                    query='''
{table}
{query_project}
| where TimeGenerated >= datetime({start})
| where TimeGenerated <= datetime({end})
| extend extendedProps = parse_json(ExtendedProperties)
| extend CompromisedEntity = tostring(extendedProps['Compromised Host'])
| project-away extendedProps
| where SystemAlertId == \'{system_alert_id}\'
{add_query_items}
''',
                    description='Retrieves an alert by alert Id',
                    data_source='security_alert',
                    data_families=[DataFamily.SecurityAlert],
                    data_environments=[DataEnvironment.LogAnalytics],
                    optional_params=['add_query_items']))

_add_query(KqlQuery(name='list_related_alerts',
                    query='''
let src_host = \'{host_name}\';
let src_acct = \'{account_name}\';
let src_proc = \'{process_name}\';
{table}
{query_project}
| where {subscription_filter}
| where TimeGenerated >= datetime({start})
| where TimeGenerated <= datetime({end})
| extend Computer = src_host
| extend src_hostname = tostring(split(src_host, '.')[0])
| extend src_accountname = iif(src_acct contains '\\\\',
                               tostring(split(src_acct, '\\\\')[-1]),
                               tostring(split(src_acct, '@')[0]))
| extend src_procname = tostring(split(src_proc, \'{path_separator}\')[-1])
| extend host_match = iif(isnotempty(src_host) and
    (Entities has src_hostname or Entities has src_host
     or ExtendedProperties has src_hostname
     or ExtendedProperties has src_host), true, false)
| extend acct_match = iif(isnotempty(src_acct)
     and (Entities has src_accountname or Entities has src_acct
     or ExtendedProperties has src_accountname
     or ExtendedProperties has src_acct), true, false)
| extend proc_match = iif(isnotempty(src_acct)
     and (Entities has src_procname or Entities has src_proc
     or ExtendedProperties has src_procname
     or ExtendedProperties has src_proc), true, false)
| where host_match or acct_match or proc_match
{add_query_items}
''',
                    description='Retrieves list of alerts with a common host, acount or process',
                    data_source='security_alert',
                    data_families=[DataFamily.SecurityAlert],
                    data_environments=[DataEnvironment.LogAnalytics],
                    optional_params=['process_name', 'account_name', 'add_query_items']))

_add_query(KqlQuery(name='list_related_ip_alerts',
                    query='''
let src_ips = \'{source_ip_list}\';
let src_ips_arr = split(src_ips, ',');
let IP_table = toscalar(range idx from 0 to array_length(src_ips_arr) - 1 step 1
| extend ip = trim(@'\\s*', tostring(src_ips_arr[idx]))
| project ip
| distinct ip
| summarize makeset(ip) );
let ip_extract = materialize(
{table}
{query_project}
| where {subscription_filter}
| where TimeGenerated >= datetime({start})
| where TimeGenerated <= datetime({end})
| project SystemAlertId, ExtendedProperties, Entities
| extend source_ips_str = extract("\\"Source IPs\\": \\"([^\\"]+)\\"", 1, ExtendedProperties)
| extend source_ips_1 = iif(isnotempty(source_ips_str), split(source_ips_str, ','), dynamic([]))
| extend source_ips_2 = extract_all("\\"Address\\": \\"([^\\"]+)\\"", dynamic([1]), Entities)
| mvexpand alert_ip_1 = source_ips_1 to typeof(string), alert_ip_2 = source_ips_2 to typeof(string)
| where isnotempty(alert_ip_1) or isnotempty(alert_ip_2)
| where alert_ip_1 in (IP_table) or alert_ip_2 in (IP_table)
| extend matching_ips = case(isnotempty(alert_ip_1) and isnotempty(alert_ip_2), strcat(alert_ip_1, ',', alert_ip_2),
                             isnotempty(alert_ip_1), alert_ip_1,
                             isnotempty(alert_ip_2), alert_ip_2,
                             '')
| extend MatchingIps = split(matching_ips, ',')
| project-away source_ips_str, source_ips_1, source_ips_2, alert_ip_1, alert_ip_2, matching_ips
);
{table}
{query_project}
| where {subscription_filter}
| where TimeGenerated >= datetime({start})
| where TimeGenerated <= datetime({end})
| join (ip_extract) on SystemAlertId
{add_query_items}
''',
                    description='Retrieves list of alerts with a common IP Address',
                    data_source='security_alert',
                    data_families=[DataFamily.SecurityAlert],
                    data_environments=[DataEnvironment.LogAnalytics],
                    optional_params=['add_query_items']))

_add_query(KqlQuery(name='get_process_tree',
                    query='''
let start = datetime({start});
let end = datetime({end});
let sourceProcessId = \'{process_id}\';
let sourceLogonId = \'{logon_session_id}\';
let sourceProcess =
materialize(
    {table}
    {query_project}
    | where {subscription_filter}
    | where {host_filter_eq}
    | where TimeGenerated >= start
    | where TimeGenerated <= end
    | where SubjectLogonId == sourceLogonId
    | where NewProcessId == sourceProcessId
    | where NewProcessName =~ \'{process_name}\'
    | extend NodeRole = 'source', Level = 0
    | top 1 by TimeCreatedUtc desc nulls last);
let sourceTimeCreatedUtc = toscalar(sourceProcess | project TimeCreatedUtc);
let sourceParentProcessId = toscalar(sourceProcess | project ProcessId);
let system_session_id = toscalar(sourceProcess
    | extend sys_session = iff(NewProcessName contains '/', '-1', '0x3e7')
    | project sys_session );
let parentProcess = // Parent Process
materialize(
    {table}
    {query_project}
    | where {subscription_filter}
    | where {host_filter_eq}
    | where TimeGenerated >= start - time(1d)
    | where TimeGenerated <= end
    | where TimeGenerated <= sourceTimeCreatedUtc
    | where (SubjectLogonId == sourceLogonId or TargetLogonId == sourceLogonId)
    | where NewProcessId == sourceParentProcessId
    | extend NodeRole = 'parent', Level = 1
    | top 1 by TimeCreatedUtc desc nulls last);
let parentLogonId = toscalar(sourceProcess | project SubjectLogonId);
let parentTimeCreated = toscalar(sourceProcess | project TimeCreatedUtc);
let childProcesses = // Child Process
materialize(
    {table}
    {query_project}
    | where {subscription_filter}
    | where {host_filter_eq}
    | where TimeGenerated >= start
    | where TimeGenerated <= end
    | where SubjectLogonId == sourceLogonId
    | where ProcessId == sourceProcessId
    | extend NodeRole = 'child', Level = 1);

sourceProcess
| union (parentProcess)
| union (childProcesses)
| union
(
    // GrandParent Process (we ignore this if this is the system logonId)
    {table}
    {query_project}
    | where {subscription_filter}
    | where {host_filter_eq}
    | where TimeGenerated >= start - time(1d)
    | where TimeGenerated <= end
    | where TimeGenerated <= parentTimeCreated
    | where (SubjectLogonId == parentLogonId or TargetLogonId == parentLogonId)
    | extend NodeRole = 'parent', Level = 2
    | join (parentProcess | project ProcessId) on $left.NewProcessId == $right.ProcessId
)
| union
(
    // GrandChild Process (we ignore this if this is the system logonId)
    {table}
    {query_project}
    | where {subscription_filter}
    | where {host_filter_eq}
    | where TimeGenerated >= start
    | where TimeGenerated <= end
    | where SubjectLogonId == sourceLogonId and SubjectLogonId != system_session_id
    | extend NodeRole = 'child', Level = 2
    | join (childProcesses | project NewProcessId) on $left.ProcessId == $right.NewProcessId
)
| union
(
    // Sibling Process
    {table}
    {query_project}
    | where {subscription_filter}
    | where {host_filter_eq}
    | where TimeGenerated >= start
    | where TimeGenerated <= end
    | where SubjectLogonId == sourceLogonId
    | where ProcessId == sourceParentProcessId
    | where NewProcessId != sourceProcessId
    | extend NodeRole = 'sibling', Level = 1
)
{add_query_items}
''',
                    description='Retrieves process tree for a process.',
                    data_source='process_create',
                    data_families=[DataFamily.WindowsSecurity,
                                   DataFamily.LinuxSecurity],
                    data_environments=[DataEnvironment.LogAnalytics],
                    optional_params=['add_query_items']))

_add_query(KqlQuery(name='list_processes',
                    query='''
let start = datetime({start});
let end = datetime({end});
{table}
{query_project}
| where {subscription_filter}
| where {host_filter_eq}
| where TimeGenerated >= start
| where TimeGenerated <= end
{add_query_items}
''',
                    description='Retrieves processes for a host.',
                    data_source='process_create',
                    data_families=[DataFamily.WindowsSecurity,
                                   DataFamily.LinuxSecurity],
                    data_environments=[DataEnvironment.LogAnalytics],
                    optional_params=['add_query_items']))

_add_query(KqlQuery(name='get_process_parent',
                    query='''
let start = datetime({start});
let end = datetime({end});
let sourceProcessId = \'{process_id}\';
let sourceLogonId = \'{logon_session_id}\';
let sourceProcess =
materialize(
    {table}
    {query_project}
    | where {subscription_filter}
    | where {host_filter_eq}
    | where TimeGenerated >= start
    | where TimeGenerated <= end
    | where SubjectLogonId == sourceLogonId
    | where NewProcessId == sourceProcessId
    | where NewProcessName =~ \'{process_name}\'
    | extend NodeRole = 'source', Level = 0
    | top 1 by TimeCreatedUtc desc nulls last);
let sourceTimeCreatedUtc = toscalar(sourceProcess | project TimeCreatedUtc );
let sourceParentProcessId = toscalar(sourceProcess | project ProcessId);
// Parent Process
{table}
{query_project}
| where {subscription_filter}
| where {host_filter_eq}
| where TimeGenerated >= start - time(2h)
| where TimeGenerated <= end
| where TimeGenerated <= sourceTimeCreatedUtc
| where (SubjectLogonId == sourceLogonId or TargetLogonId == sourceLogonId)
| where NewProcessId == sourceParentProcessId
| where NewProcessId == sourceParentProcessId
| extend NodeRole = 'parent', Level = 1
| top 1 by TimeCreatedUtc desc nulls last);
{add_query_items}
''',
                    description='Retrieves the parent process of a process process',
                    data_source='process_create',
                    data_families=[DataFamily.WindowsSecurity,
                                   DataFamily.LinuxSecurity],
                    data_environments=[DataEnvironment.LogAnalytics],
                    optional_params=['add_query_items']))

_add_query(KqlQuery(name='list_hosts_matching_commandline',
                    query='''
{table}
{query_project}
| where {subscription_filter}
| where {host_filter_neq}
| where TimeGenerated >= datetime({start})
| where TimeGenerated <= datetime({end})
| where NewProcessName endswith \'{process_name}\'
| where CommandLine =~ \'{commandline}\'
{add_query_items}
''',
                    description='Retrieves processes on other hosts with matching commandline',
                    data_source='process_create',
                    data_families=[DataFamily.WindowsSecurity,
                                   DataFamily.LinuxSecurity],
                    data_environments=[DataEnvironment.LogAnalytics],
                    optional_params=['add_query_items']))

_add_query(KqlQuery(name='list_processes_in_session',
                    query='''
{table}
{query_project}
| where {subscription_filter}
| where {host_filter_eq}
| where TimeGenerated >= datetime({start})
| where TimeGenerated <= datetime({end})
| where SubjectLogonId == \'{logon_session_id}\'
| extend processName = tostring(split(NewProcessName, \'{path_separator}\')[-1])
| extend commandlineparts = arraylength(split(CommandLine, ' '))
| extend commandlinelen = strlen(CommandLine)
{add_query_items}
''',
                    description='Retrieves all processes on the host for a logon session',
                    data_source='process_create',
                    data_families=[DataFamily.WindowsSecurity,
                                   DataFamily.LinuxSecurity],
                    data_environments=[DataEnvironment.LogAnalytics],
                    optional_params=['add_query_items']))

_add_query(KqlQuery(name='get_host_logon',
                    query='''
{table}
{query_project}
| where {subscription_filter}
| where {host_filter_eq}
| where TimeGenerated >= datetime({start})
| where TimeGenerated <= datetime({end})
| where TargetLogonId == \'{logon_session_id}\'
{add_query_items}
''',
                    description='Retrieves the logon event for the session id on the host.',
                    data_source='account_logon',
                    data_families=[DataFamily.WindowsSecurity,
                                   DataFamily.LinuxSecurity],
                    data_environments=[DataEnvironment.LogAnalytics],
                    optional_params=['add_query_items']))

_add_query(KqlQuery(name='list_host_logons',
                    query='''
{table}
{query_project}
| where {subscription_filter}
| where {host_filter_eq}
| where TimeGenerated >= datetime({start})
| where TimeGenerated <= datetime({end})
{add_query_items}
''',
                    description='Retrieves the logon events on the host.',
                    data_source='account_logon',
                    data_families=[DataFamily.WindowsSecurity,
                                   DataFamily.LinuxSecurity],
                    data_environments=[DataEnvironment.LogAnalytics],
                    optional_params=['add_query_items']))

_add_query(KqlQuery(name='list_host_logon_failures',
                    query='''
{table}
{query_project}
| where {subscription_filter}
| where {host_filter_eq}
| where TimeGenerated >= datetime({start})
| where TimeGenerated <= datetime({end})
{add_query_items}
''',
                    description='Retrieves the logon failure events on the host.',
                    data_source='account_logon_fail',
                    data_families=[DataFamily.WindowsSecurity,
                                   DataFamily.LinuxSecurity],
                    data_environments=[DataEnvironment.LogAnalytics],
                    optional_params=['add_query_items']))



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Miscellaneous helper methods for Jupyter Notebooks."""
import re
import sys
from IPython.core.display import display, HTML

import pandas as pd

from .. _version import VERSION

__version__ = VERSION
__author__ = 'Ian Hellen'


def export(func):
    """Decorate function or class to export to __all__."""
    mod = sys.modules[func.__module__]
    if hasattr(mod, '__all__'):
        mod.__all__.append(func.__name__)
    else:
        mod.__all__ = [func.__name__]
    return func


@export
def string_empty(string):
    """Return True if the input string is None or whitespace."""
    return (string is None) or not (string and string.strip())


@export
def is_not_empty(test_object):
    """Return True if the test_object is not None or empty."""
    if test_object:
        if isinstance(test_object, str):
            if test_object.strip():
                # test_object is not None AND myString is not empty or blank
                return True
            return False
        return True
    return False


# Toggle Code Cell Contents
_TOGGLE_CODE_STR = '''
<form action="javascript:code_toggle()">
    <input type="submit" id="toggleButton" value="Show/Hide Code">
</form>
'''

_TOGGLE_CODE_PREPARE_STR = '''
    <script>
    function code_toggle() {
        if ($('div.cell.code_cell.rendered.selected div.input').css('display')!='none'){
            $('div.cell.code_cell.rendered.selected div.input').hide();
        } else {
            $('div.cell.code_cell.rendered.selected div.input').show();
        }
    }
    </script>

'''
display(HTML(_TOGGLE_CODE_PREPARE_STR))


@export
def toggle_code():
    """Display a toggle button to hide/reveal code cell."""
    display(HTML(_TOGGLE_CODE_STR))


# String escapes
@export
def escape_windows_path(str_path):
    """Escape backslash characters in a string."""
    if is_not_empty(str_path):
        return str_path.replace('\\', '\\\\')
    return str_path


@export
def unescape_windows_path(str_path):
    """Remove escaping from backslash characters in a string."""
    if is_not_empty(str_path):
        return str_path.replace('\\\\', '\\')
    return str_path


_PD_INSTALLED_VERSION = tuple(pd.__version__.split('.'))
_PD_VER_23 = ('0', '23', '0')


@export
def pd_version_23() -> bool:
    """Return True if pandas version 0.23.0 or later is installed."""
    return _PD_INSTALLED_VERSION >= _PD_VER_23


@export
def get_nb_query_param(nb_url_search: str, param: str) -> str:
    """
    Get a url query parameter from the search string.

    Arguments:
        nb_url_search {str} -- The URL search string
        param {str} -- The parameter name to search for

    Returns:
        value of the query string parameter or None if not found.

    """
    qs_regex = r'[\\?&]{param}=(?P<val>[^&#]*)'.format(param=param)
    query_string_match = re.search(qs_regex, nb_url_search)
    if query_string_match:
        return query_string_match['val']
    return None


@export
def get_nb_query_params(nb_url_search: str) -> dict:
    """
    Get the url query parameters from the search string.

    Arguments:
        nb_url_search {str} -- The URL search string

    Returns:
        dictionary of the query string parameters.

    """
    nb_params = {}
    query_string_match = re.search(r'\?(?P<qs>[^#]+)#?', nb_url_search)
    if query_string_match:
        for param in query_string_match['qs'].split('&'):
            if '=' in param:
                nb_params[param.split('=')[0]] = param.split('=')[1]
    return nb_params


@export
def get_notebook_query_string():
    """Execute javascript to publish notebook query string as python variable."""
    HTML('''
    <script type="text/javascript">
        IPython.notebook.kernel.execute(
            "nb_query_string='".concat(window.location.search).concat("'"));
    </script>
    ''')



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""KQL Helper functions."""
import sys
from functools import partial

import pandas as pd
from IPython import get_ipython
from Kqlmagic import results

from . query_builtin_queries import query_definitions
# pylint: disable=locally-disabled, W0611
# (list_queries not used here but want to bring in into module namespace)
from . query_mgr import (replace_prov_query_params, list_queries,
                         clean_kql_query, query_help, print_kql)
# pylint: enable=locally-disabled, W0611
from . utility import export
from .. _version import VERSION

__version__ = VERSION
__author__ = 'Ian Hellen'

# pylint: disable=locally-disabled, C0103
_kql_magic_loaded = False
_ip = get_ipython()
# pylint: enable=locally-disabled, C0103


@export
def load_kql_magic():
    """Load KqlMagic if not loaded."""
    # KqlMagic
    if not _is_kqlmagic_loaded():
        print('Please wait. Loading Kqlmagic extension...')
        get_ipython().run_line_magic('reload_ext', 'Kqlmagic')
        # get_ipython().run_line_magic('config', 'Kqlmagic.auto_dataframe=True')
        if not _is_kqlmagic_loaded():
            raise EnvironmentError('Kqlmagic did not load correctly.')


def _is_kqlmagic_loaded() -> bool:
    """Return true if kql magic is loaded."""
    # pylint: disable=locally-disabled, C0103, W0603
    global _kql_magic_loaded
    # pylint: enable=locally-disabled, C0103, W0603

    if _ip is not None:
        _kql_magic_loaded = _ip.find_magic('kql') is not None

    return _kql_magic_loaded


@export
def exec_query(query_name: str, **kwargs) -> (pd.DataFrame, results.ResultSet):
    """
    Execute kql query with optional parameters and return a Dataframe.

    Use list_queries() to see the current set).
    Use query_help(query_name) to view the query and expected paramaters
    Arguments:
        query_name {string}: the name of query to run
        kwargs: additional replacable paramters for the query
            {string:bool} kql_result=True - return (DataFrame, KqlResultSet)
                    tuple.
                    kql_result=False - return DataFrame only (default)
            {string:[QueryParamProvider]} - for the key 'provs'
                (or alias 'providers')
                this should be a collection of objects that
                implement QueryParamProvider (from which query
                parameters can be extracted).
                OR
            {string:value pairs} -- custom parameter list
                (override default values and values extracted
                from QueryParamProviders).
    Returns:
        dataframe {pd.DataFrame}: if kql_result == False (default).
        (dataframe, ResultSet): tuple of dataframe and Kql ResultSet
            if kql_result==True (pass this as a kw argument).

    """
    if 'kql_result' in kwargs:
        kql_result = kwargs.pop('kql_result')
    else:
        kql_result = False

    replaced_query = replace_prov_query_params(query_name=query_name, **kwargs)
    replaced_query = clean_kql_query(replaced_query)
    if replaced_query:
        result = _ip.run_cell_magic('kql', line='', cell=replaced_query)
        if result is not None and result.completion_query_info['StatusCode'] == 0:
            data_frame = result.to_dataframe()
            if result.is_partial_table:
                print("Warning - query returned partial results.")
            # Did user want both dataframe and ResultSet
            if not kql_result:
                return data_frame

            return data_frame, result

        print("Warning - query did not complete successfully.")
        print("Kql ResultSet returned - check  \'completion_query_info\' property.")
        return result


@export
def show_filled_query(query_name: str, **kwargs) -> (pd.DataFrame, results.ResultSet):
    """
    Print the kql query with replaced parameter values.

    Use list_queries() to see the current set).
    Use query_help(query_name) to view the query and expected paramaters
    Arguments:
        query_name {string}: the name of query to run
        kwargs: additional replacable paramters for the query
            {string:bool} kql_result=True - return (DataFrame, KqlResultSet)
                    tuple.
                    kql_result=False - return DataFrame only (default)
            {string:[QueryParamProvider]} - for the key 'provs'
                (or alias 'providers')
                this should be a collection of objects that
                implement QueryParamProvider (from which query
                parameters can be extracted).
                OR
            {string:value pairs} -- custom parameter list
                (override default values and values extracted
                from QueryParamProviders).
    Returns:
        replaced_query {str}: the query with substituted parameters.

    """
    replaced_query = replace_prov_query_params(query_name=query_name, **kwargs)
    print_kql(replaced_query)
    return replaced_query


def _add_queries_to_module(module_name):
    """Add queries to the module as callable methods."""
    if module_name not in sys.modules:
        raise LookupError(f'Module {module_name} was not found sys.modules')
    for query_name in query_definitions:
        module = sys.modules[module_name]
        query_func = partial(exec_query, query_name=query_name)
        query_func.__doc__ = exec_query.__doc__
        setattr(module, query_name, query_func)


# Add all queries defined in builtin queries module as functions
if __name__ != '__main__':
    _add_queries_to_module(__name__)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Base64unpack test class."""
import unittest
import os
from os import path
import pandas as pd

from .. msticpy.sectools import base64unpack as b64


_test_data_folders = [d for d, _, _ in os.walk(os.getcwd()) if d.endswith('/tests/testdata')]
if len(_test_data_folders) == 1:
    _TEST_DATA = _test_data_folders[0]
else:
    _TEST_DATA = './tests/testdata'


class TestB64Unpack(unittest.TestCase):
    """Unit test class."""
    def test_archive_string(self):
        try:
            input_txt = None
            FILE_NAME = path.join(_TEST_DATA, 'b64text_inzip.txt')
            with open(FILE_NAME, 'r') as f_handle:
                input_txt = f_handle.read()

            self.assertIsNotNone(input_txt)

            result_str, result_df = b64.unpack_items(input_string=input_txt, trace=True)
            print(result_str)
            # assert result_df.shape == (2, 12)
            self.assertIsNotNone(result_str)
            self.assertIsNotNone(result_df)

        except FileNotFoundError as ex:
            self.fail(msg='Exception {}'.format(str(ex)))

    def test_nested_archive(self):
        try:
            input_txt = None
            FILE_NAME = path.join(_TEST_DATA, 'base64msg.txt')
            with open(FILE_NAME, 'r') as f_handle:
                input_txt = f_handle.read()

            self.assertIsNotNone(input_txt)

            result_str, result_df = b64.unpack_items(input_string=input_txt, trace=True)
            assert result_df.shape == (8, 12)
            self.assertIsNotNone(result_str)
            self.assertIsNotNone(result_df)

        except FileNotFoundError as ex:
            self.fail(msg='Exception {}'.format(str(ex)))

    def test_nested_archive_df(self):
        try:
            input_txt = None
            FILE_NAME = path.join(_TEST_DATA, 'base64msg.txt')
            with open(FILE_NAME, 'r') as f_handle:
                input_txt = f_handle.read()

            self.assertIsNotNone(input_txt)
            # create datframe for input with 2 rows and add same data to both
            input_df = pd.DataFrame(data=['a', 'b'], columns=['input'], index=[0, 1])
            input_df['input'] = input_txt
            result_df = b64.unpack_items(data=input_df, column='input', trace=True)
            # we should get 2x the rows as the previous test (since data is duplicated)
            # plus 2 added columns
            assert result_df.shape == (16, 14)
            self.assertIsNotNone(result_df)

        except FileNotFoundError as ex:
            self.fail(msg='Exception {}'.format(str(ex)))


if __name__ == '__main__':
    unittest.main()
    print('bye')



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Setup script for msticpy."""

import re

import setuptools

INSTALL_REQUIRES = ['matplotlib>=3.0.0',
                    'bokeh>=1.0.2',
                    'setuptools>=40.6.2',
                    'attrs>=18.2.0',
                    'pandas>=0.23.0',
                    'requests>=2.20.1',
                    'networkx>=2.2',
                    'numpy>=1.12.0',
                    'urllib3>=1.24.0',
                    'ipywidgets>=7.4.2',
                    'ipython>=7.1.1',
                    'Kqlmagic>=0.1.90',
                    'scikit_learn>=0.20.2',
                    'maxminddb_geolite2>=2018.0',
                    'typing>=3.6.6']


# pylint: disable=locally-disabled, C0103
with open("README.md", "r") as fh:
    long_description = fh.read()
# pylint: enable=locally-disabled, C0103

with open("msticpy/_version.py", "r") as fd:
    __version__ = re.search(r'^VERSION\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE).group(1)

setuptools.setup(
    name="msticpy",
    version=__version__,
    author="Ian Hellen",
    author_email="ianhelle@microsoft.com",
    description="MSTIC Security Tools",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://https://github.com/ianhelle/msyticpy",
    python_requires='>=3.6',
    packages=setuptools.find_packages(exclude=['*.tests']),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=INSTALL_REQUIRES
)
